"""
Module to run tests on arcoadd
"""
import os

import pytest
import numpy as np

from pypeit import msgs
from pypeit.tests.tstutils import cooked_required
from IPython import embed


@cooked_required
def test_cooked_version():
    # TODO: Use pygit2 to both add the branch name when constructing
    # Cooked and here to check that the branch is correct?

    # Load up the version
    v_file = os.path.join(os.getenv('PYPEIT_DEV'), 'Cooked', 'version')
    with open(v_file) as f:
        tmp = f.readlines()
    assert tmp[-1].strip() == '0.13.3'





""" Uber object for calibration images, e.g. arc, flat """

import inspect

import os
import numpy as np


from pypeit import msgs
from pypeit.par import pypeitpar
from pypeit.images import combineimage

from IPython import embed


class CalibrationImage(object):
    """
    Class to generate (and hold) a combined calibration image from a list of input images or
    simply to hold a previously generated calibration image.

    Args:
        spectrograph (:class:`pypeit.spectrographs.spectrograph.Spectrograph`):
            Spectrograph used to take the data.
        det (:obj:`int`, optional):
            The 1-indexed detector number to process.
        proc_par (:class:`pypeit.par.pypeitpar.ProcessImagesPar` or None):
            Parameters that dictate the processing of the images.  See
            :class:`pypeit.par.pypeitpar.ProcessImagesPar` for the
            defaults.

        files (list, optional):
            List of filenames to be combined

    Attributes:
        pypeitImage (:class:`pypeit.images.pypeitimage.PypeItImage`):
        file_list (list):
            List of files to process
        process_steps (list):
            List of processing steps to be used

    """
    def __init__(self, spectrograph, det, proc_par, files=None):

        # Required items
        self.spectrograph = spectrograph
        self.det = det

        # Assign the internal list of files
        self._set_files(files)

        # Required parameters
        if proc_par is not None:
            if not isinstance(proc_par, pypeitpar.ProcessImagesPar):
                msgs.error('Provided ParSet for must be type ProcessImagesPar.')
        self.proc_par = proc_par  # This must be named this way as it is frequently a child

        # Process steps
        self.process_steps = []

        # Standard output
        self.pypeitImage = None

    @property
    def nfiles(self):
        """
        The number of calibration files
        """
        return len(self.file_list)

    def _set_files(self, files, check=False):
        """
        Assign the provided files to :attr:`files`.

        Args:
            files (None, :obj:`str`, :obj:`list`):
                The files to process.
            check (:obj:`bool`, optional):
                Check that the files exist on disk.

        Raises:
            PypeItError:
                Raised if the input objects have the wrong type.
        """
        if files is None:
            self.file_list = []
        elif isinstance(files, str):
            self.file_list = [files]
        elif isinstance(files, list):
            if not np.all([isinstance(f, str) for f in files]):
                msgs.error('File list elements must be strings.')
            self.file_list = files
        else:
            msgs.error('Provides files must be None, a string name, or a list of strings.')

        if check:
            self._check_files()

    def _check_files(self):
        """
        Check that files in :attr:`files` exist.

        Raises:
            PypeItError:
                Raised if any of the files don't exist.
        """
        for f in self.file_list:
            if not os.path.isfile(f):
                msgs.error('{0} does not exist!'.format(f))

    def build_image(self, bias=None, bpm=None, ignore_saturation=True):
        """
        Load, process and combine the images for a Calibration

        Args:
            bias (np.ndarray, optional):
                Bias image
            bpm (np.ndarray, optional):
                Bad pixel mask
            ignore_saturation (bool, optional):
                If True, turn off the saturation flag in the individual images before stacking
                This avoids having such values set to 0 which for certain images (e.g. flat calibrations)
                has unintended consequences.

        Returns:
            PypeItImage:

        """
        combineImage = combineimage.CombineImage(self.spectrograph, self.det, self.proc_par, self.file_list)
        self.pypeitImage = combineImage.run(self.process_steps, bias, bpm=bpm, ignore_saturation=ignore_saturation)
        # Return
        return self.pypeitImage


    def __repr__(self):
        return ('<{:s}: nfiles={}, steps={}>'.format(
            self.__class__.__name__, self.nfiles, self.process_steps))








"""
Module to run tests on SciImgStack class
Requires files in Development suite
"""
import os

import pytest
import glob
import numpy as np

from pypeit.tests.tstutils import dev_suite_required, cooked_required
from pypeit.tests.tstutils import load_kast_blue_masters
from pypeit.spectrographs.util import load_spectrograph
from pypeit.images import scienceimage
from pypeit.images import rawimage
from pypeit.images import processrawimage
from pypeit.core import procimg


def data_path(filename):
    data_dir = os.path.join(os.path.dirname(__file__), 'files')
    return os.path.join(data_dir, filename)

kast_blue = load_spectrograph('shane_kast_blue')
kast_par = kast_blue.default_pypeit_par()
keck_nires = load_spectrograph('keck_nires')
nires_par = keck_nires.default_pypeit_par()

@pytest.fixture
@dev_suite_required
def shane_kast_blue_sci_files():
    return [os.path.join(os.getenv('PYPEIT_DEV'), 'RAW_DATA', 'shane_kast_blue', '600_4310_d55',
                         ifile) for ifile in ['b27.fits.gz', 'b28.fits.gz']]

@pytest.fixture
@dev_suite_required
def nires_sci_files():
    return [os.path.join(os.getenv('PYPEIT_DEV'), 'RAW_DATA', 'keck_nires', 'NIRES', ifile)
            for ifile in ['s180604_0089.fits.gz', 's180604_0092.fits.gz']]

@pytest.fixture
@dev_suite_required
def nires_bg_files():
    return [os.path.join(os.getenv('PYPEIT_DEV'), 'RAW_DATA', 'keck_nires', 'NIRES', ifile)
            for ifile in ['s180604_0090.fits.gz', 's180604_0091.fits.gz']]


def test_standard_instantiate():
    """
    Simple instantiate
    """
    # Empty
    bpm = np.zeros((100,100))
    det = 1
    sciImg = scienceimage.ScienceImage(kast_blue, det, kast_par['scienceframe']['process'],
                                       np.zeros_like(bpm),
                                       np.zeros_like(bpm),
                                       bpm)


@cooked_required
def test_instantiate_from_one(shane_kast_blue_sci_files):
    """
    Run on a single science frame
    """
    #
    det = 1
    # Load calibrations
    pixelflat = load_kast_blue_masters(pixflat=True)[0]
    bpm = kast_blue.empty_bpm(shane_kast_blue_sci_files[0], det)
    # Process steps
    bias = None
    par = kast_par['scienceframe']['process']
    process_steps = procimg.init_process_steps(bias, par)
    process_steps += ['trim', 'apply_gain', 'orient']
    process_steps += ['flatten']
    process_steps += ['extras']
    process_steps += ['crmask']
    # Load
    rawImage = rawimage.RawImage(shane_kast_blue_sci_files[0], kast_blue, det)
    processRawImage = processrawimage.ProcessRawImage(rawImage,
                                                      kast_par['scienceframe']['process'])
    pypeItImage = processRawImage.process(process_steps, pixel_flat=pixelflat)
    # Do it
    sciImg = scienceimage.ScienceImage(kast_blue, det, kast_par['scienceframe']['process'],
                                       pypeItImage.image, pypeItImage.ivar, bpm)


@cooked_required
def test_from_list(shane_kast_blue_sci_files):
    """
    Run on two frames
    """
    det = 1
    # Load calibrations
    pixelflat = load_kast_blue_masters(pixflat=True)[0]
    bpm = kast_blue.empty_bpm(shane_kast_blue_sci_files[0], det)
    # Do it
    sciImg = scienceimage.build_from_file_list(kast_blue, det,
                                               kast_par['scienceframe']['process'],
                                               bpm, shane_kast_blue_sci_files,
                                               None, pixelflat)
    # Test
    assert isinstance(sciImg, scienceimage.ScienceImage)


@dev_suite_required
def test_proc_diff(nires_sci_files, nires_bg_files):
    """
    Run on near-IR frames
    """
    # Setup
    det = 1
    bpm = np.zeros((2048,1024))
    pixelflat = np.ones_like(bpm)

    # Sci image
    sciImg = scienceimage.build_from_file_list(keck_nires, det,
                                               nires_par['scienceframe']['process'], bpm,
                                               nires_sci_files, None, pixelflat)
    # Bg image
    bgImg = scienceimage.build_from_file_list(keck_nires, det,
                                              nires_par['scienceframe']['process'], bpm,
                                              nires_bg_files, None, pixelflat)
    # Difference
    sciImg = sciImg - bgImg
    # Test
    assert isinstance(sciImg, scienceimage.ScienceImage)



""" Module for fluxing routines
"""
import glob
import numpy as np
import os
import scipy

from pkg_resources import resource_filename

from astropy import units
from astropy import constants
from astropy import coordinates
from astropy.table import Table, Column
from astropy.io import ascii
from astropy.io import fits

from matplotlib import pyplot as plt

from linetools.spectra.xspectrum1d import XSpectrum1D
from astropy import table

from pypeit import msgs
from pypeit import utils
from pypeit import debugger
from pypeit.wavemodel import conv2res
from pypeit.core import pydl, load, save, coadd
from pypeit.spectrographs.util import load_spectrograph
from pypeit import specobjs

from IPython import embed

TINY = 1e-15
MAGFUNC_MAX = 25.0
MAGFUNC_MIN = -25.0
SN2_MAX = (20.0) ** 2
PYPEIT_FLUX_SCALE = 1e-17


def find_standard_file(ra, dec, toler=20.*units.arcmin, check=False):
    """
    Find a match for the input file to one of the archived
    standard star files (hopefully).  Priority is by order of search.

    Args:
        ra (str):
            Object right-ascension in hh:mm:ss string format (e.g.,
            '05:06:36.6').
        dec (str):
            Object declination in dd:mm:ss string format (e.g.,
            52:52:01.0')
        toler (:class:`astropy.units.quantity.Quantity`, optional):
            Tolerance on matching archived standards to input.  Expected
            to be in arcmin.
        check (:obj:`bool`, optional):
            If True, the routine will only check to see if a standard
            star exists within the input ra, dec, and toler range.

    Returns:
        dict, bool: If check is True, return True or False depending on
        if the object is matched to a library standard star.  If check
        is False and no match is found, return None.  Otherwise, return
        a dictionary with the matching standard star with the following
        meta data:

            - 'file': str -- Filename table
            - 'name': str -- Star name
            - 'ra': str -- RA(J2000)
            - 'dec': str -- DEC(J2000)

    """
    # Priority
    std_sets = ['xshooter', 'calspec', 'esofil']

    # SkyCoord
    try:
        ra, dec = float(ra), float(dec)
        obj_coord = coordinates.SkyCoord(ra, dec, unit=(units.deg, units.deg))
    except:
        obj_coord = coordinates.SkyCoord(ra, dec, unit=(units.hourangle, units.deg))

    # Loop on standard sets
    closest = dict(sep=999 * units.deg)

    for sset in std_sets:
        path = resource_filename('pypeit', os.path.join('data', 'standards', sset))
        star_file =  os.path.join(path, '{0}_info.txt'.format(sset))
        if not os.path.isfile(star_file):
            msgs.warn('File does not exist!: {0}'.format(star_file))
            continue

        star_tbl = Table.read(star_file, comment='#', format='ascii')
        star_coords = coordinates.SkyCoord(star_tbl['RA_2000'], star_tbl['DEC_2000'],
                                           unit=(units.hourangle, units.deg))
        idx, d2d, d3d = coordinates.match_coordinates_sky(obj_coord, star_coords, nthneighbor=1)

        if d2d < toler:
            if check:
                # Found one so return
                return True

            # Generate a dict
            _idx = int(idx)
            std_dict = dict(cal_file=os.path.join(path, star_tbl[_idx]['File']),
                            name=star_tbl[_idx]['Name'],
                            std_ra=star_tbl[_idx]['RA_2000'],
                            std_dec=star_tbl[_idx]['DEC_2000'])

            if not os.path.isfile(star_file):
                # TODO: Error or warn?
                msgs.error("No standard star file found: {:s}".format(star_file))

            # TODO: Does this need to be globbed? Why isn't the file
            # name exact?
            fil = glob.glob(std_dict['cal_file'] + '*')
            if len(fil) == 0:
                # TODO: Error or warn?
                msgs.error("No standard star file: {:s}".format(std_dict['cal_file']))

            fil = fil[0]
            msgs.info("Loading standard star file: {:s}".format(fil))
            # TODO: Put this stuf in a method, like `read_standard`
            if sset == 'xshooter':
                # TODO let's add the star_mag here and get a uniform set of tags in the std_dict
                std_spec = Table.read(fil, format='ascii')
                std_dict['std_source'] = sset
                std_dict['wave'] = std_spec['col1'] * units.AA
                std_dict['flux'] = std_spec['col2'] / PYPEIT_FLUX_SCALE * \
                                   units.erg / units.s / units.cm ** 2 / units.AA
            elif sset == 'calspec':
                std_dict['std_source'] = sset
                std_spec = fits.open(fil)[1].data
                std_dict['wave'] = std_spec['WAVELENGTH'] * units.AA
                std_dict['flux'] = std_spec['FLUX'] / PYPEIT_FLUX_SCALE \
                                   * units.erg / units.s / units.cm ** 2 / units.AA
            elif sset == 'esofil':
                # TODO let's add the star_mag here and get a uniform set of tags in the std_dict
                std_spec = Table.read(fil, format='ascii')
                std_dict['std_source'] = sset
                std_dict['wave'] = std_spec['col1'] * units.AA
                std_dict['flux'] = std_spec['col2']*1e-16/PYPEIT_FLUX_SCALE * \
                                   units.erg / units.s / units.cm ** 2 / units.AA
            else:
                msgs.error('Do not know how to parse {0} file.'.format(sset))
            msgs.info("Fluxes are flambda, normalized to 1e-17")
            return std_dict

        # Save closest found so far
        imind2d = np.argmin(d2d)
        mind2d = d2d[imind2d]
        if mind2d < closest['sep']:
            closest['sep'] = mind2d
            # TODO: Is this right? Do we need to use the imind2d from
            # above?
            closest.update(dict(name=star_tbl[int(idx)]['Name'],
                                ra=star_tbl[int(idx)]['RA_2000'],
                                dec=star_tbl[int(idx)]['DEC_2000']))

    # Standard star not found
    if check:
        return False

    msgs.error("No standard star was found within a tolerance of {:g}".format(toler) + msgs.newline()
               + "Closest standard was {:s} at separation {:g}".format(closest['name'], closest['sep'].to('arcmin')))

    return None

def stellar_model(V, sptype):
    """
    Parse Kurucz SED given T and g.  Also convert absolute/apparent
    magnitudes

    Parameters
    ----------
    V: float
        Apparent magnitude of the telluric star
    sptype: str
        Spectral type of the telluric star

    Returns
    -------
    loglam: ndarray
        log wavelengths
    flux: ndarray
        SED f_lambda (cgs units, I think, probably per Ang)
    """

    # Grab telluric star parameters
    # log(g) of the Sun
    logg_sol = np.log10(6.67259e-8) + np.log10(1.989e33) - 2.0 * np.log10(6.96e10)

    # Load Schmidt-Kaler (1982) table
    sk82_file = resource_filename('pypeit', 'data/standards/kurucz93/schmidt-kaler_table.txt')
    sk82_tab = ascii.read(sk82_file, names=('Sp', 'logTeff', 'Teff', '(B-V)_0', 'M_V', 'B.C.', 'M_bol', 'L/L_sol'))

    # TODO, currently this only works on select stellar types. Add ability to interpolate across types.
    # Match input type.
    mti = np.where(sptype == sk82_tab['Sp'])[0]
    if len(mti) != 1:
        raise ValueError('Not ready to interpolate yet.')

    # Calculate final quantities
    # Relation between radius, temp, and bolometric luminosity
    logR = 0.2 * (42.26 - sk82_tab['M_bol'][mti[0]] - 10.0 * sk82_tab['logTeff'][mti[0]])

    # Mass-bolometric luminosity relation from schimdt-kaler p28 valid for M_bol < 7.5
    logM = 0.46 - 0.10 * sk82_tab['M_bol'][mti[0]]
    logg = logM - 2.0 * logR + logg_sol
    M_V = sk82_tab['M_V'][mti[0]]
    Teff = sk82_tab['Teff'][mti[0]]

    # Flux factor (absolute/apparent V mag)
    # Define constants
    parsec = constants.pc.cgs  # 3.086e18
    R_sol = constants.R_sun.cgs  # 6.96e10

    # Distance modulus
    logd = 0.2 * (V - M_V) + 1.0
    D = parsec * 10. ** logd
    R = R_sol * 10. ** logR

    # Factor converts the kurucz surface flux densities to flux observed on Earth
    flux_factor = (R / D.value) ** 2

    # Grab closest T in Kurucz SEDs
    T1 = 3000. + np.arange(28) * 250
    T2 = 10000. + np.arange(6) * 500
    T3 = 13000. + np.arange(22) * 1000
    T4 = 35000. + np.arange(7) * 2500
    Tk = np.concatenate([T1, T2, T3, T4])
    indT = np.argmin(np.abs(Tk - Teff))

    # Grab closest g in Kurucz SEDs
    loggk = np.arange(11) * 0.5
    indg = np.argmin(np.abs(loggk - logg))

    # Grab Kurucz filename
    std_file = resource_filename('pypeit', '/data/standards/kurucz93/kp00/kp00_{:d}.fits.gz'.format(int(Tk[indT])))
    std = Table.read(std_file)

    # Grab specific spectrum
    loglam = np.array(np.log10(std['WAVELENGTH']))
    gdict = {0: 'g00', 1: 'g05', 2: 'g10', 3: 'g15', 4: 'g20',
             5: 'g25', 6: 'g30', 7: 'g35', 8: 'g40', 9: 'g45',
             10: 'g50'}
    flux = std[gdict[indg]]

    # scale the model to the V-band magnitude
    star_lam = 10 ** loglam
    star_flux = flux.data * flux_factor
    # Generate a dict matching the output of find_standard_file
    std_dict = dict(cal_file='KuruczTelluricModel', name=sptype, Vmag=V, std_ra=None, std_dec=None)
    std_dict['std_source'] = 'KuruczTelluricModel'
    std_dict['wave'] = star_lam * units.AA
    std_dict['flux'] = star_flux / PYPEIT_FLUX_SCALE * units.erg / units.s / units.cm ** 2 / units.AA

    return std_dict

def get_standard_spectrum(star_type=None, star_mag=None, ra=None, dec=None):
    '''
    Get the standard spetrum using given information of your standard/telluric star.

    Args:
        star_type: str
            Spectral type of your standard/telluric star
        star_mag: float
            Apparent magnitude of the telluric star
        ra: str
            Standard right-ascension in hh:mm:ss string format (e.g.,'05:06:36.6').
        dec: str
            Object declination in dd:mm:ss string format (e.g., 52:52:01.0')

    Returns:
        dict: Dictionary containing the information you provided and the
        standard/telluric spectrum.
    '''
    # Create star model
    if (ra is not None) and (dec is not None) and (star_mag is None) and (star_type is None):
        # Pull star spectral model from archive
        msgs.info("Getting archival standard spectrum")
        # Grab closest standard within a tolerance
        std_dict = find_standard_file(ra, dec)

    elif (star_mag is not None) and (star_type is not None):
        ## using vega spectrum
        if 'A0' in star_type:
            msgs.info('Getting vega spectrum')
            ## Vega model from TSPECTOOL
            vega_file = resource_filename('pypeit', '/data/standards/vega_tspectool_vacuum.dat')
            vega_data = Table.read(vega_file, comment='#', format='ascii')
            std_dict = dict(cal_file='vega_tspectool_vacuum', name=star_type, Vmag=star_mag, std_ra=ra, std_dec=dec)
            std_dict['std_source'] = 'VEGA'
            std_dict['wave'] = vega_data['col1'] * units.AA

            # vega is V=0.03
            std_dict['flux'] = vega_data['col2'] * 10**(0.4*(0.03-star_mag)) / PYPEIT_FLUX_SCALE * \
                               units.erg / units.s / units.cm ** 2 / units.AA
        ## using Kurucz stellar model
        else:
            # Create star spectral model
            msgs.info("Getting kurucz+93 stellar model")
            std_dict = stellar_model(star_mag, star_type)
            std_dict['std_ra'] = ra
            std_dict['std_dec'] = dec
    else:
        msgs.error('Insufficient information provided for fluxing. '
                   'Either the coordinates of the standard or a stellar type and magnitude are needed.')

    return std_dict


def load_extinction_data(longitude, latitude, toler=5. * units.deg):
    """
    Find the best extinction file to use, based on longitude and latitude
    Loads it and returns a Table

    Parameters
    ----------
    toler : Angle, optional
        Tolerance for matching detector to site (5 deg)

    Returns
    -------
    ext_file : Table
        astropy Table containing the 'wavelength', 'extinct' data for AM=1.
    """
    # Mosaic coord
    mosaic_coord = coordinates.SkyCoord(longitude, latitude, frame='gcrs', unit=units.deg)
    # Read list
    extinct_path = resource_filename('pypeit', '/data/extinction/')
    extinct_summ = extinct_path + 'README'
    extinct_files = Table.read(extinct_summ, comment='#', format='ascii')
    # Coords
    ext_coord = coordinates.SkyCoord(extinct_files['Lon'], extinct_files['Lat'], frame='gcrs',
                                     unit=units.deg)
    # Match
    idx, d2d, d3d = coordinates.match_coordinates_sky(mosaic_coord, ext_coord, nthneighbor=1)
    if d2d < toler:
        extinct_file = extinct_files[int(idx)]['File']
        msgs.info("Using {:s} for extinction corrections.".format(extinct_file))
    else:
        msgs.warn("No file found for extinction corrections.  Applying none")
        msgs.warn("You should generate a site-specific file")
        return None
    # Read
    extinct = Table.read(extinct_path + extinct_file, comment='#', format='ascii',
                         names=('iwave', 'mag_ext'))
    wave = Column(np.array(extinct['iwave']) * units.AA, name='wave')
    extinct.add_column(wave)
    # Return
    return extinct[['wave', 'mag_ext']]

def extinction_correction(wave, airmass, extinct):
    """
    Derive extinction correction
    Based on algorithm in LowRedux (long_extinct)

    Parameters
    ----------
    wave : ndarray
        Wavelengths for interpolation. Should be sorted Assumes
        Angstroms
    airmass : float
        Airmass
    extinct : Table
        Table of extinction values

    Returns
    -------
    flux_corr : ndarray
        Flux corrections at the input wavelengths
    """
    # Checks
    if airmass < 1.:
        msgs.error("Bad airmass value in extinction_correction")
    # Interpolate
    f_mag_ext = scipy.interpolate.interp1d(extinct['wave'],extinct['mag_ext'], bounds_error=False, fill_value=0.)
    mag_ext = f_mag_ext(wave)#.to('AA').value)

    # Deal with outside wavelengths
    gdv = np.where(mag_ext > 0.)[0]

    if len(gdv) == 0:
        msgs.warn("No valid extinction data available at this wavelength range. Extinction correction not applied")
    elif gdv[0] != 0:  # Low wavelengths
        mag_ext[0:gdv[0]] = mag_ext[gdv[0]]
        msgs.warn("Extrapolating at low wavelengths using last valid value")
    elif gdv[-1] != (mag_ext.size - 1):  # High wavelengths
        mag_ext[gdv[-1] + 1:] = mag_ext[gdv[-1]]
        msgs.warn("Extrapolating at high wavelengths using last valid value")
    else:
        msgs.info("Extinction data covered the whole spectra. Correct it!")
    # Evaluate
    flux_corr = 10.0 ** (0.4 * mag_ext * airmass)
    # Return
    return flux_corr


### Routines for standard sensfunc started from here
def find_standard(specobj_list):
    """
    Take the median boxcar and then the max object as the standard

    Parameters
    ----------
    specobj_list : list

    Returns
    -------
    mxix : int
        Index of the standard star

    """
    # Repackage as necessary (some backwards compatability)
    # Do it
    medfx = []
    for indx, spobj in enumerate(specobj_list):
        if spobj is None:
            medfx.append(0.)
        else:
            medfx.append(np.median(spobj.BOX_COUNTS))
    try:
        mxix = np.argmax(np.array(medfx))
    except:
        debugger.set_trace()
    msgs.info("Putative standard star {} has a median boxcar count of {}".format(specobj_list[mxix],
                                                                                 np.max(medfx)))
    # Return
    return mxix

#def apply_standard_sens(spec_obj, sens_dict, airmass, exptime, extinct_correct=True, telluric_correct = False,
#                        longitude=None, latitude=None):
#    """ Apply the sensitivity function to the data
#    We also correct for extinction.
#
#    Parameters
#    ----------
#    spec_obj : dict
#        SpecObj
#    sens_dict : dict
#        Sens Function dict
#    airmass : float
#        Airmass
#    exptime : float
#        Exposure time in seconds
#    longitude : float
#        longitude in degree for observatory
#    latitude: float
#        latitude in degree for observatory. Used for extinction
#        correction
#    """

def sensfunc(wave, counts, counts_ivar, counts_mask, exptime, airmass, std_dict, longitude, latitude, ech_orders=None,
             mask_abs_lines=True, telluric=False, polyorder=4, balm_mask_wid=5., nresln=20., resolution=3000.,
             trans_thresh=0.9,polycorrect=True, debug=False):
    """
    Function to generate the sensitivity function. This function fits
    a bspline to the 2.5*log10(flux_std/flux_counts). The break
    points spacing, which determines the scale of variation of the
    sensitivity function is determined by the nresln parameter. This
    code can work in different regimes, but NOTE THAT TELLURIC MODE
    IS DEPRECATED, use telluric.sensfunc_telluric instead

    Args:
        wave (ndarray):
            Wavelength of the star. Shape (nspec,) or (nspec, norders)
        counts (ndarray):
            Flux (in counts) of the star. Shape (nspec,) or (nspec, norders)
        counts_ivar (ndarray):
            Inverse variance of the star counts. Shape (nspec,) or (nspec, norders)
        counts_mask (ndarray):
            Good pixel mask for the counts. Shape (nspec,) or (nspec, norders)
        exptime (float):
            Exposure time in seconds
        airmass (float):
            Airmass
        std_dict (dict):
            Dictionary containing information about the standard star returned by flux_calib.get_standard_spectrum
        longitude (float):
            Telescope longitude, used for extinction correction.
        latitude (float):
            Telescope latitude, used for extinction correction
        telluric (bool):
            If True attempts to fit telluric absorption. This feature is deprecated, as one should instead
            use telluric.sensfunc_telluric. Default=False
        ech_orders (ndarray, int):
            If passed the echelle orders will be added to the meta_table. ech_orders must be a numpy array of integers
            with the shape (norders,) giving the order numbers
        mask_abs_lines (bool):
            If True, mask stellar absorption lines before fitting sensitivity function. Default = True
        balm_mask_wid (float):
            Parameter describing the width of the mask for or stellar absorption lines (i.e. mask_abs_lines=True). A region
            equal to balm_mask_wid*resln is masked where resln is the estimate for the spectral resolution in pixels
            per resolution element.
        polycorrect (bool):
            Whether you want to interpolate the sensfunc with polynomial in the stellar absortion line regions before
            fitting with the bspline
        nresln (float):
            Parameter governing the spacing of the bspline breakpoints. default = 20.0
        resolution (float):
            Expected resolution of the standard star spectrum. This should probably be determined from the grating, but is
            currently hard wired. default=3000.0
        trans_thresh (float):
            Parameter for selecting telluric regions which are masked. Locations below this transmission value are masked.
            If you have significant telluric absorption you should be using telluric.sensnfunc_telluric. default = 0.9

    Returns:
        Tuple: Returns:

            - meta_table (astropy.Table) -- Table containing meta data
              for the sensitivity function
            - out_table (astropy.Table) -- Table containing the
              sensitivity function

    """

    wave_arr, counts_arr, ivar_arr, mask_arr, nspec, norders = utils.spec_atleast_2d(wave, counts, counts_ivar, counts_mask)
    sensfunc = np.zeros_like(wave_arr)
    mask_sens = np.ones_like(mask_arr)
    wave_min = np.zeros(norders)
    wave_max = np.zeros(norders)

    for iord in range(norders):
        sensfunc[:, iord], mask_sens[:, iord] = sensfunc_eval(
            wave_arr[:,iord], counts_arr[:,iord], ivar_arr[:,iord], mask_arr[:,iord], exptime, airmass, std_dict,
            longitude, latitude, mask_abs_lines=mask_abs_lines, telluric=telluric, polyorder=polyorder,
            balm_mask_wid=balm_mask_wid, nresln=nresln, resolution=resolution, trans_thresh=trans_thresh,
            polycorrect=polycorrect, debug=debug)
        wave_min[iord] = wave_arr[wave_arr[:,iord] > 1.0, iord].min()
        wave_max[iord] = wave_arr[wave_arr[:,iord] > 1.0, iord].max()

    # Allocate the meta parameter table, ext=1
    meta_table = table.Table(meta={'name': 'Parameter Values'})
    meta_table['EXPTIME'] = [exptime]
    meta_table['AIRMASS'] = [airmass]
    meta_table['STD_RA'] = [std_dict['std_ra']]
    meta_table['STD_DEC'] = [std_dict['std_dec']]
    meta_table['STD_NAME'] = [std_dict['name']]
    meta_table['CAL_FILE'] = [std_dict['cal_file']]
    if ech_orders is not None:
        meta_table['ECH_ORDERS'] = [ech_orders]
    # Allocate the output table, ext=2
    out_table = table.Table(meta={'name': 'Sensitivity Function'})
    # These are transposed because we need to store them in an astropy table, with number of rows = norders
    out_table['WAVE'] = wave_arr.T
    out_table['SENSFUNC'] = sensfunc.T
    out_table['MASK_SENS'] = mask_sens.T
    out_table['WAVE_MIN'] = wave_min
    out_table['WAVE_MAX'] = wave_max

    return meta_table, out_table



# JFH TODO This code needs to be cleaned up. The telluric option should probably be removed. Logic is not easy to follow.
def sensfunc_eval(wave, counts, counts_ivar, counts_mask, exptime, airmass, std_dict, longitude, latitude,
                  mask_abs_lines=True, telluric=False, polyorder=4, balm_mask_wid=5., nresln=20., resolution=3000.,
                  trans_thresh=0.9,polycorrect=True, debug=False):

    """

    Function to generate the sensitivity function. This function fits
    a bspline to the 2.5*log10(flux_std/flux_counts). The break
    points spacing, which determines the scale of variation of the
    sensitivity function is determined by the nresln parameter. This
    code can work in different regimes, but NOTE THAT TELLURIC MODE
    IS DEPRECATED, use telluric.sensfunc_telluric instead.

        - If telluric=False, a sensfunc is generated by fitting a
          bspline to the using nresln=20.0 and masking out telluric
          regions.

        - If telluric=True, sensfunc is a pixelized sensfunc (not
          smooth) for correcting both throughput and telluric lines.
          if you set polycorrect=True, the sensfunc in the Hydrogen
          recombination line region (often seen in star spectra) will
          be replaced by a smoothed polynomial function.

    Args:
        wave (ndarray):
            Wavelength of the star. Shape (nspec,)
        counts (ndarray):
            Flux (in counts) of the star. Shape (nspec,)
        counts_ivar (ndarray):
            Inverse variance of the star counts. Shape (nspec,)
        counts_mask (ndarray):
            Good pixel mask for the counts.
        exptime (float):
            Exposure time in seconds
        airmass (float):
            Airmass
        std_dict (dict):
            Dictionary containing information about the standard star returned by flux_calib.get_standard_spectrum
        longitude (float):
            Telescope longitude, used for extinction correction.
        latitude (float):
            Telescope latitude, used for extinction correction
        telluric (bool):
            If True attempts to fit telluric absorption. This feature is deprecated, as one should instead
            use telluric.sensfunc_telluric. Default=False
        mask_abs_lines (bool):
            If True, mask stellar absorption lines before fitting sensitivity function. Default = True
        balm_mask_wid (float):
            Parameter describing the width of the mask for or stellar absorption lines (i.e. mask_abs_lines=True). A region
            equal to balm_mask_wid*resln is masked where resln is the estimate for the spectral resolution in pixels
            per resolution element.
        polycorrect: bool
            Whether you want to interpolate the sensfunc with polynomial in the stellar absortion line regions before
            fitting with the bspline
        nresln (float):
            Parameter governing the spacing of the bspline breakpoints. default = 20.0
        resolution (float):
            Expected resolution of the standard star spectrum. This should probably be determined from the grating, but is
            currently hard wired. default=3000.0
        trans_thresh (float):
            Parameter for selecting telluric regions which are masked. Locations below this transmission value are masked.
            If you have significant telluric absorption you should be using telluric.sensnfunc_telluric. default = 0.9

    Returns:
        tuple: Returns:

            - sensfunc (ndarray) -- Sensitivity function with same shape as wave (nspec,)
            - mask_sens (ndarray, bool) -- Good pixel mask for sensitivity function with same shape as wave (nspec,)

    """
    # Create copy of the arrays to avoid modification and convert to
    # electrons / s
    wave_star = wave.copy()
    flux_star = counts.copy() / exptime
    ivar_star = counts_ivar.copy() * exptime ** 2

    # Extinction correction
    msgs.info("Applying extinction correction")
    extinct = load_extinction_data(longitude,latitude)
    ext_corr = extinction_correction(wave * units.AA, airmass, extinct)
    # Correct for extinction
    flux_star = flux_star * ext_corr
    ivar_star = ivar_star / ext_corr ** 2
    mask_star = counts_mask

    # Interpolate the standard star onto the current set of observed wavelengths
    flux_true = scipy.interpolate.interp1d(std_dict['wave'], std_dict['flux'],bounds_error=False, fill_value='extrapolate')(wave_star)
    # Do we need to extrapolate? TODO Replace with a model or a grey body?
    if np.min(flux_true) <= 0.:
        msgs.warn('Your spectrum extends beyond calibrated standard star, extrapolating the spectra with polynomial.')
        mask_model = flux_true <= 0
        msk_poly, poly_coeff = utils.robust_polyfit_djs(std_dict['wave'].value, std_dict['flux'].value,8,function='polynomial',
                                                    invvar=None, guesses=None, maxiter=50, inmask=None, \
                                                    lower=3.0, upper=3.0, maxdev=None, maxrej=3, groupdim=None,
                                                    groupsize=None,groupbadpix=False, grow=0, sticky=True, use_mad=True)
        star_poly = utils.func_val(poly_coeff, wave_star.value, 'polynomial')
        #flux_true[mask_model] = star_poly[mask_model]
        flux_true = star_poly.copy()
        if debug:
            plt.plot(std_dict['wave'], std_dict['flux'],'bo',label='Raw Star Model')
            plt.plot(std_dict['wave'],  utils.func_val(poly_coeff, std_dict['wave'].value, 'polynomial'), 'k-',label='robust_poly_fit')
            plt.plot(wave_star,flux_true,'r-',label='Your Final Star Model used for sensfunc')
            plt.show()

    # Get masks from observed star spectrum. True = Good pixels
    mask_bad, mask_balm, mask_tell = get_mask(wave_star, flux_star, ivar_star, mask_star, mask_abs_lines=mask_abs_lines,
                                              mask_telluric=True, balm_mask_wid=balm_mask_wid, trans_thresh=trans_thresh)

    # Get sensfunc
    #LBLRTM = False
    #if LBLRTM:
    #    # sensfunc = lblrtm_sensfunc() ???
    #    msgs.develop('fluxing and telluric correction based on LBLRTM model is under developing.')
    #else:
    sensfunc, mask_sens = standard_sensfunc(
        wave_star, flux_star, ivar_star, mask_bad, flux_true, mask_balm=mask_balm,
        mask_tell=mask_tell, maxiter=35, upper=3.0, lower=3.0, polyorder=polyorder,
        balm_mask_wid=balm_mask_wid, nresln=nresln,telluric=telluric, resolution=resolution,
        polycorrect=polycorrect, debug=debug, show_QA=False)

    if debug:
        plt.plot(wave_star.value[mask_sens], flux_true[mask_sens], color='k',lw=2,label='Reference Star')
        plt.plot(wave_star.value[mask_sens], flux_star[mask_sens]*sensfunc[mask_sens], color='r',label='Fluxed Observed Star')
        plt.xlabel(r'Wavelength [$\AA$]')
        plt.ylabel('Flux [erg/s/cm2/Ang.]')
        plt.legend(fancybox=True, shadow=True)
        plt.show()


    return sensfunc, mask_sens




def get_mask(wave_star,flux_star, ivar_star, mask_star, mask_abs_lines=True, mask_telluric=True, balm_mask_wid=10., trans_thresh=0.9):
    '''
    Get a couple of masks from your observed standard spectrum.

    Args:
      wave_star: numpy array
        wavelength array of your spectrum
      flux_star: numpy array
        flux array of your spectrum
      ivar_star:
        ivar array of your spectrum
      mask_star: bool
        whether you need to mask Hydrogen recombination line region. If False, the returned msk_star are all good.
      mask_tell: bool
        whether you need to mask telluric region. If False, the returned msk_tell are all good.
      trans_thresh: float
        parameter for selecting telluric regions.

    Returns:
      msk_bad: bool type numpy array
        mask for bad pixels.
      msk_star: bool type numpy array
        mask for recombination lines in star spectrum.
      msk_tell: bool type numpy array
        mask for telluric regions.
    '''

    # Mask (True = good pixels)
    # mask for recombination lines
    mask_balm = np.ones_like(flux_star).astype(bool)
    # mask for telluric regions
    mask_tell = np.ones_like(flux_star).astype(bool)

    # masking bad entries
    msgs.info(" Masking bad pixels")
    mask_bad = mask_star.copy()
    mask_bad[ivar_star <= 0.] = False
    mask_bad[flux_star <= 0.] = False
    # Mask edges
    msgs.info(" Masking edges")
    mask_bad[[0, -1]] = False
    # Mask Atm. cutoff
    msgs.info(" Masking Below the atmospheric cutoff")
    atms_cutoff = wave_star <= 3000.0
    mask_bad[atms_cutoff] = False

    # TODO JFH replace with mask_star_lines from telluric.py
    if mask_abs_lines:
        # Mask Balmer, Paschen, Brackett, and Pfund recombination lines
        msgs.info("Masking recombination lines:")
        # Mask Balmer
        msgs.info(" Masking Balmer")
        lines_balm = np.array([3836.4, 3969.6, 3890.1, 4102.8, 4102.8, 4341.6, 4862.7, 5407.0,
                               6564.6, 8224.8, 8239.2])
        for line_balm in lines_balm:
            ibalm = np.abs(wave_star - line_balm) <= balm_mask_wid
            mask_balm[ibalm] = False
        # Mask Paschen
        msgs.info(" Masking Paschen")
        # air wavelengths from:
        # https://www.subarutelescope.org/Science/Resources/lines/hi.html
        lines_pasc = np.array([8203.6, 8440.3, 8469.6, 8504.8, 8547.7, 8600.8, 8667.4, 8752.9,
                               8865.2, 9017.4, 9229.0, 9546.0, 10049.4, 10938.1,
                               12818.1, 18751.0])
        for line_pasc in lines_pasc:
            ipasc = np.abs(wave_star - line_pasc) <= balm_mask_wid
            mask_balm[ipasc] = False
        # Mask Brackett
        msgs.info(" Masking Brackett")
        # air wavelengths from:
        # https://www.subarutelescope.org/Science/Resources/lines/hi.html
        lines_brac = np.array([14584.0, 18174.0, 19446.0, 21655.0,26252.0, 40512.0])
        for line_brac in lines_brac:
            ibrac = np.abs(wave_star - line_brac) <= balm_mask_wid
            mask_balm[ibrac] = False
        # Mask Pfund
        msgs.info(" Masking Pfund")
        # air wavelengths from:
        # https://www.subarutelescope.org/Science/Resources/lines/hi.html
        lines_pfund = np.array([22788.0, 32961.0, 37395.0, 46525.0,74578.0])
        for line_pfund in lines_pfund:
            ipfund = np.abs(wave_star - line_pfund) <= balm_mask_wid
            mask_balm[ipfund] = False

    if mask_telluric:
        ## Mask telluric region in the optical
        tell_opt = np.any([((wave_star >= 6270.00) & (wave_star <= 6290.00)), # H2O
                       ((wave_star >= 6850.00) & (wave_star <= 6960.00)), #O2 telluric band
                       ((wave_star >= 7580.00) & (wave_star <= 7750.00)), #O2 telluric band
                       ((wave_star >= 7160.00) & (wave_star <= 7340.00)), #H2O
                       ((wave_star >= 8150.00) & (wave_star <= 8250.00))],axis=0) #H2O
        mask_tell[tell_opt] = False
        ## Mask near-infrared telluric region
        if np.max(wave_star)>9100.0:
            # ToDo: should use the specific atmosphere transmission after FBD get the grid.
            ## Read atmosphere transmission
            #
            #if watervp <1.5:
            #    skytrans_file = resource_filename('pypeit', '/data/skisim/'+'mktrans_zm_10_10.dat')
            #elif (watervp>=1.5 and watervp<2.3):
            #    skytrans_file = resource_filename('pypeit', '/data/skisim/'+'mktrans_zm_16_10.dat')
            #elif (watervp>=2.3 and watervp<4.0):
            #    skytrans_file = resource_filename('pypeit', '/data/skisim/' + 'mktrans_zm_30_10.dat')
            #else:
            #    skytrans_file = resource_filename('pypeit', '/data/skisim/' + 'mktrans_zm_50_10.dat')
            #
            skytrans_file = resource_filename('pypeit', '/data/skisim/' + 'mktrans_zm_10_10.dat')
            skytrans = ascii.read(skytrans_file)
            wave_trans, trans = skytrans['wave'].data*10000.0, skytrans['trans'].data
            trans_use = (wave_trans>=np.min(wave_star)-100.0) & (wave_trans<=np.max(wave_star)+100.0)
            # Estimate the resolution of your spectra.
            # I assumed 3 pixels per resolution. This gives an approximate right resolution at the middle point.
            resolution = np.median(wave_star) / np.median(wave_star - np.roll(wave_star, 1)) / 3
            trans_convolved, px_sigma, px_bin = conv2res(wave_trans[trans_use], trans[trans_use], resolution,
                                                         central_wl='midpt', debug=False)
            trans_final = scipy.interpolate.interp1d(wave_trans[trans_use], trans_convolved,bounds_error=False,
                                                     fill_value='extrapolate')(wave_star)
            tell_nir = (trans_final<trans_thresh) & (wave_star>9100.0)
            mask_tell[tell_nir] = False
        else:
            msgs.info('Your spectrum is bluer than 9100A, only optical telluric regions are masked.')

    return mask_bad, mask_balm, mask_tell

def standard_sensfunc(wave, flux, ivar, mask_bad, flux_std, mask_balm=None, mask_tell=None,
                 maxiter=35, upper=2.0, lower=2.0, polyorder=5, balm_mask_wid=50., nresln=20., telluric=True,
                 resolution=2700., polycorrect=True, debug=False, show_QA=False):
    """
    Generate a sensitivity function based on observed flux and standard spectrum.

    Parameters
    ----------
    wave : ndarray
      wavelength as observed
    flux : ndarray
      counts/s as observed
    ivar : ndarray
      inverse variance
    flux_std : Quantity array
      standard star true flux (erg/s/cm^2/A)
    msk_bad : ndarray
      mask for bad pixels. True is good.
    msk_star: ndarray
      mask for hydrogen recombination lines. True is good.
    msk_tell:ndarray
      mask for telluric regions. True is good.
    maxiter : integer
      maximum number of iterations for polynomial fit
    upper : integer
      number of sigma for rejection in polynomial
    lower : integer
      number of sigma for rejection in polynomial
    polyorder : integer
      order of polynomial fit
    balm_mask_wid: float
      in units of angstrom
      Mask parameter for Balmer absorption. A region equal to
      balm_mask_wid is masked.
    resolution: integer/float.
      spectra resolution
      This paramters should be removed in the future. The resolution should be estimated from spectra directly.
    debug : bool
      if True shows some dubugging plots

    Returns
    -------
    sensfunc
    """
    # Create copy of the arrays to avoid modification
    wave_obs = wave.copy()
    flux_obs = flux.copy()
    ivar_obs = ivar.copy()
    # preparing arrays
    if np.any(np.invert(np.isfinite(ivar_obs))):
        msgs.warn("NaN are present in the inverse variance")

    # check masks
    if mask_tell is None:
        mask_tell = np.ones_like(wave_obs,dtype=bool)
    if mask_balm is None:
        mask_balm = np.ones_like(wave_obs, dtype=bool)

    # Removing outliers
    # Calculate log of flux_obs setting a floor at TINY
    logflux_obs = 2.5 * np.log10(np.maximum(flux_obs, TINY))
    # Set a fix value for the variance of logflux
    logivar_obs = np.ones_like(logflux_obs) * (10.0 ** 2)
    # Calculate log of flux_std model setting a floor at TINY
    logflux_std = 2.5 * np.log10(np.maximum(flux_std, TINY))
    # Calculate ratio setting a floor at MAGFUNC_MIN and a ceiling at
    # MAGFUNC_MAX
    magfunc = logflux_std - logflux_obs
    magfunc = np.maximum(np.minimum(magfunc, MAGFUNC_MAX), MAGFUNC_MIN)
    msk_magfunc = (magfunc < 0.99 * MAGFUNC_MAX) & (magfunc > 0.99 * MAGFUNC_MIN)

    # Define two new masks, True is good and False is masked pixel
    # mask for all bad pixels on sensfunc
    masktot = mask_bad & msk_magfunc & np.isfinite(ivar_obs) & np.isfinite(logflux_obs) & np.isfinite(logflux_std)
    logivar_obs[np.invert(masktot)] = 0.0
    # mask used for polynomial fit
    msk_fit_sens = masktot & mask_tell & mask_balm

    # Polynomial fitting to derive a smooth sensfunc (i.e. without telluric)
    _, poly_coeff = utils.robust_polyfit_djs(wave_obs[msk_fit_sens], magfunc[msk_fit_sens], polyorder,
                                             function='polynomial', invvar=None, guesses=None, maxiter=maxiter,
                                             inmask=None, lower=lower, upper=upper, maxdev=None,
                                             maxrej=None, groupdim=None, groupsize=None, groupbadpix=False,
                                             grow=0, sticky=True, use_mad=True)
    magfunc_poly = utils.func_val(poly_coeff, wave_obs, 'polynomial')

    # Polynomial corrections on Hydrogen Recombination lines
    if ((sum(msk_fit_sens) > 0.5 * len(msk_fit_sens)) & polycorrect):
        ## Only correct Hydrogen Recombination lines with polyfit in the telluric free region
        balmer_clean = np.zeros_like(wave_obs, dtype=bool)
        # Commented out the bluest recombination lines since they are weak for spectroscopic standard stars.
        #836.4, 3969.6, 3890.1, 4102.8, 4102.8, 4341.6, 4862.7,   \
        lines_hydrogen = np.array([5407.0, 6564.6, 8224.8, 8239.2, 8203.6, 8440.3, 8469.6, 8504.8, 8547.7, 8600.8, \
                                   8667.4, 8752.9, 8865.2, 9017.4, 9229.0, 10049.4, 10938.1, 12818.1, 21655.0])
        for line_hydrogen in lines_hydrogen:
            ihydrogen = np.abs(wave_obs - line_hydrogen) <= balm_mask_wid
            balmer_clean[ihydrogen] = True
        msk_clean = ((balmer_clean) | (magfunc == MAGFUNC_MAX) | (magfunc == MAGFUNC_MIN)) & \
                    (magfunc_poly > MAGFUNC_MIN) & (magfunc_poly < MAGFUNC_MAX)
        magfunc[msk_clean] = magfunc_poly[msk_clean]
        msk_badpix = np.isfinite(ivar_obs) & (ivar_obs > 0)
        magfunc[np.invert(msk_badpix)] = magfunc_poly[np.invert(msk_badpix)]
    else:
        ## if half more than half of your spectrum is masked (or polycorrect=False) then do not correct it with polyfit
        msgs.warn('No polynomial corrections performed on Hydrogen Recombination line regions')

    if not telluric:
        # Apply mask to ivar
        #logivar_obs[~msk_fit_sens] = 0.

        # ToDo
        # Compute an effective resolution for the standard. This could be improved
        # to setup an array of breakpoints based on the resolution. At the
        # moment we are using only one number
        msgs.work("Should pull resolution from arc line analysis")
        msgs.work("At the moment the resolution is taken as the PixelScale")
        msgs.work("This needs to be changed!")
        std_pix = np.median(np.abs(wave_obs - np.roll(wave_obs, 1)))
        std_res = np.median(wave_obs/resolution) # median resolution in units of Angstrom.
        #std_res = std_pix
        #resln = std_res
        if (nresln * std_res) < std_pix:
            msgs.warn("Bspline breakpoints spacing shoud be larger than 1pixel")
            msgs.warn("Changing input nresln to fix this")
            nresln = std_res / std_pix

        # Fit magfunc with bspline
        kwargs_bspline = {'bkspace': std_res * nresln}
        kwargs_reject = {'maxrej': 5}
        msgs.info("Initialize bspline for flux calibration")
        init_bspline = pydl.bspline(wave_obs, bkspace=kwargs_bspline['bkspace'])
        fullbkpt = init_bspline.breakpoints

        # TESTING turning off masking for now
        # remove masked regions from breakpoints
        msk_obs = np.ones_like(wave_obs).astype(bool)
        msk_obs[np.invert(masktot)] = False
        msk_bkpt = scipy.interpolate.interp1d(wave_obs, msk_obs, kind='nearest', fill_value='extrapolate')(fullbkpt)
        init_breakpoints = fullbkpt[msk_bkpt > 0.999]

        # init_breakpoints = fullbkpt
        msgs.info("Bspline fit on magfunc. ")
        bset1, bmask = pydl.iterfit(wave_obs, magfunc, invvar=logivar_obs, inmask=msk_fit_sens, upper=upper, lower=lower,
                                    fullbkpt=init_breakpoints, maxiter=maxiter, kwargs_bspline=kwargs_bspline,
                                    kwargs_reject=kwargs_reject)
        logfit1, _ = bset1.value(wave_obs)
        logfit_bkpt, _ = bset1.value(init_breakpoints)

        if debug:
            # Check for calibration
            plt.figure(1)
            plt.plot(wave_obs, magfunc, drawstyle='steps-mid', color='black', label='magfunc')
            plt.plot(wave_obs, logfit1, color='cornflowerblue', label='logfit1')
            plt.plot(wave_obs[np.invert(msk_fit_sens)], magfunc[np.invert(msk_fit_sens)], '+', color='red', markersize=5.0,
                     label='masked magfunc')
            plt.plot(wave_obs[np.invert(msk_fit_sens)], logfit1[np.invert(msk_fit_sens)], '+', color='red', markersize=5.0,
                     label='masked logfit1')
            plt.plot(init_breakpoints, logfit_bkpt, '.', color='green', markersize=4.0, label='breakpoints')
            plt.plot(init_breakpoints, np.interp(init_breakpoints, wave_obs, magfunc), '.', color='green',
                     markersize=4.0,
                     label='breakpoints')
            plt.plot(wave_obs, 1.0 / np.sqrt(logivar_obs), color='orange', label='sigma')
            plt.legend()
            plt.xlabel('Wavelength [ang]')
            plt.ylim(0.0, 1.2 * MAGFUNC_MAX)
            plt.title('1st Bspline fit')
            plt.show()
        # Create sensitivity function
        magfunc = np.maximum(np.minimum(logfit1, MAGFUNC_MAX), MAGFUNC_MIN)
        if ((sum(msk_fit_sens) > 0.5 * len(msk_fit_sens)) & polycorrect):
            msk_clean = ((magfunc == MAGFUNC_MAX) | (magfunc == MAGFUNC_MIN)) & \
                        (magfunc_poly > MAGFUNC_MIN) & (magfunc_poly<MAGFUNC_MAX)
            magfunc[msk_clean] = magfunc_poly[msk_clean]
            msk_badpix = np.isfinite(ivar_obs) & (ivar_obs>0)
            magfunc[np.invert(msk_badpix)] = magfunc_poly[np.invert(msk_badpix)]
        else:
            ## if half more than half of your spectrum is masked (or polycorrect=False) then do not correct it with polyfit
            msgs.warn('No polynomial corrections performed on Hydrogen Recombination line regions')

    # Calculate sensfunc
    sensfunc = 10.0 ** (0.4 * magfunc)

    if debug:
        plt.figure()
        magfunc_raw = logflux_std - logflux_obs
        plt.plot(wave_obs[masktot],magfunc_raw[masktot] , 'k-',lw=3,label='Raw Magfunc')
        plt.plot(wave_obs[masktot],magfunc_poly[masktot] , 'c-',lw=3,label='Polynomial Fit')
        plt.plot(wave_obs[np.invert(mask_tell)], magfunc_raw[np.invert(mask_tell)], 's',
                 color='0.7',label='Telluric Region')
        plt.plot(wave_obs[np.invert(mask_balm)], magfunc_raw[np.invert(mask_balm)], 'r+',label='Recombination Line region')
        plt.plot(wave_obs[masktot], magfunc[masktot],'b-',label='Final Magfunc')
        plt.legend(fancybox=True, shadow=True)
        plt.xlim([0.995*np.min(wave_obs[masktot]),1.005*np.max(wave_obs[masktot])])
        plt.ylim([0.,1.2*np.max(magfunc[masktot])])
        plt.show()
        plt.close()

    return sensfunc, masktot

def load_filter_file(filter):
    """
    Load a system response curve for a given filter

    Args:
        filter (str): Name of filter

    Returns:
        ndarray, ndarray: wavelength, instrument throughput

    """
    '''
    # Optical filters
    BASS_MZLS_filters = ['BASS-MZLS-{}'.format(i) for i in ['G', 'R','Z']]
    CFHT_filters = ['CFHT-{}'.format(i) for i in ['U', 'G', 'R', 'I', 'Z']]
    DECAM_filters = ['DECAM-{}'.format(i) for i in ['U', 'G', 'R', 'I', 'Z', 'Y']]
    HSC_filters = ['HSC-{}'.format(i) for i in ['G', 'R', 'I', 'Z', 'Y']]
    LSST_filters = ['LSST-{}'.format(i) for i in ['U', 'G', 'R', 'I', 'Z', 'Y']]
    PS1_filters = ['PS1-{}'.format(i) for i in ['G', 'R', 'I', 'Z', 'Y']]
    SDSS_filters = ['SDSS-{}'.format(i) for i in ['U', 'G', 'R', 'I', 'Z']]

    # NIR filters
    UKIDSS_filters = ['UKIRT-{}'.format(i) for i in ['Y', 'J', 'H', 'K']]
    VISTA_filters = ['VISTA-{}'.format(i) for i in ['Z', 'Y', 'J', 'H', 'K']]
    TMASS_filters = ['TMASS-{}'.format(i) for i in ['J', 'H', 'K']]

    # Other filters
    GAIA_filters = ['GAIA-{}'.format(i) for i in ['G', 'B', 'R']]
    GALEX_filters = ['GALEX-{}'.format(i) for i in ['F', 'N']]
    WISE_filters = ['WISE-{}'.format(i) for i in ['W1', 'W2', 'W3', 'W4']]

    allowed_options = BASS_MZLS_filters + CFHT_filters + DECAM_filters + HSC_filters \
                      + LSST_filters + PS1_filters + SDSS_filters + UKIDSS_filters\
                      + VISTA_filters + TMASS_filters + GAIA_filters + GALEX_filters + WISE_filters
    '''
    filter_file = resource_filename('pypeit', os.path.join('data', 'filters', 'filter_list.ascii'))
    tbl = Table.read(filter_file, format='ascii')

    allowed_options = tbl['filter'].data

    # Check
    if filter not in allowed_options:
        msgs.error("PypeIt is not ready for filter = {}".format(filter))

    trans_file = resource_filename('pypeit', os.path.join('data', 'filters', 'filtercurves.fits'))
    trans = fits.open(trans_file)
    wave = trans[filter].data['lam']  # Angstroms
    instr = trans[filter].data['Rlam']  # Am keeping in atmospheric terms
    keep = instr > 0.
    # Parse
    wave = wave[keep]
    instr = instr[keep]

    # Return
    return wave, instr

def scale_in_filter(xspec, scale_dict):
    """
    Scale spectra to input magnitude in given filter

    scale_dict has data model:
      - 'filter' (str): name of filter
      - 'mag' (float): magnitude
      - 'mag_type' (str, optional): type of magnitude.  Assumed 'AB'
      - 'masks' (list, optional): Wavelength ranges to mask in calculation

    Args:
        xspec (linetools.spectra.xspectrum1d.XSpectrum1D):
        scale_dict (dict):

    Returns:
        linetools.spectra.xspectrum1d.XSpectrum1D, float:  Scaled spectrum
    """
    # Parse the spectrum
    sig = xspec.sig
    gdx = sig > 0.
    wave = xspec.wavelength.value[gdx]
    flux = xspec.flux.value[gdx]

    # Mask further?
    if 'masks' in scale_dict:
        if scale_dict['masks'] is not None:
            gdp = np.ones_like(wave, dtype=bool)
            for mask in scale_dict['masks']:
                bad = (wave > mask[0]) & (wave < mask[1])
                gdp[bad] = False
            # Cut again
            wave = wave[gdp]
            flux = flux[gdp]

    if ('mag_type' in scale_dict) | (scale_dict['mag_type'] is not None):
        mag_type = scale_dict['mag_type']
    else:
        mag_type = 'AB'

    # Grab the instrument response function
    fwave, trans = load_filter_file(scale_dict['filter'])
    tfunc = scipy.interpolate.interp1d(fwave, trans, bounds_error=False, fill_value=0.)

    # Convolve
    allt = tfunc(wave)
    wflam = np.sum(flux*allt) / np.sum(allt) * PYPEIT_FLUX_SCALE * units.erg/units.s/units.cm**2/units.AA

    mean_wv = np.sum(fwave*trans)/np.sum(trans) * units.AA

    #
    if mag_type == 'AB':
        # Convert flam to AB magnitude
        fnu = wflam * mean_wv**2 / constants.c
        # Apparent AB
        AB = -2.5 * np.log10(fnu.to('erg/s/cm**2/Hz').value) - 48.6
        # Scale factor
        Dm = AB - scale_dict['mag']
        scale = 10**(Dm/2.5)
        msgs.info("Scaling spectrum by {}".format(scale))
        # Generate
        new_spec = XSpectrum1D.from_tuple((xspec.wavelength, xspec.flux*scale, xspec.sig*scale))
    else:
        msgs.error("Need a magnitude for scaling")

    return new_spec,scale




"""
Module for terminal and file logging.

.. todo::
    Why not use pythons native logging package?

"""
import sys
import os
import getpass
import glob
import textwrap
import inspect

# Imported for versioning
import scipy
import numpy
import astropy

from pypeit import __version__ #, __last_updated__
from pypeit.core.qa import close_qa

#pypeit_logger = None

# Alphabetical list of developers
developers = ['ema', 'joe', 'milvang', 'rcooke', 'thsyu', 'xavier']


class PypeItError(Exception):
    pass


class Messages:
    """
    Create coloured text for messages printed to screen.

    For further details on colours see the following example:
    http://ascii-table.com/ansi-escape-sequences.php

    Parameters
    ----------
    log : str or None
      Name of saved log file (no log will be saved if log=="")
    verbosity : int (0,1,2)
      Level of verbosity:
        0 = No output
        1 = Minimal output
        2 = All output (default)
    colors : bool
      If true, the screen output will have colors, otherwise
      normal screen output will be displayed
    """
    def __init__(self, log=None, verbosity=None, colors=True):

        # Initialize other variables
        self._defverb = 1
        if getpass.getuser() in developers:
            self._defverb = 2
        self._verbosity = self._defverb if verbosity is None else verbosity
#        self._last_updated = __last_updated__
        self._version = __version__

        # TODO: Why are these two necessary?  It would seem better to
        # provide Messages with member functions that can operate on
        # sciexp and pypeit_file instead of having them kept within the
        # object itself...
        self.sciexp = None
        self.pypeit_file = None

        # Initialize the log
        self._log = None
        self._initialize_log_file(log=log)

        # Use colors?
        self._start = None
        self._end = None
        self._black_CL = None
        self._yellow_CL = None
        self._blue_CL = None
        self._green_CL = None
        self._red_CL = None
        self._white_RD = None
        self._white_GR = None
        self._white_BK = None
        self._white_BL = None
        self._black_YL = None
        self._yellow_BK = None

        self.disablecolors()
        if colors:
            self.enablecolors()

    def _cleancolors(self, msg):
        cols = [self._end, self._start,
                self._black_CL, self._yellow_CL, self._blue_CL, self._green_CL, self._red_CL,
                self._white_RD, self._white_GR, self._white_BK, self._white_BL,
                self._black_YL, self._yellow_BK]
        for i in cols:
            msg = msg.replace(i, '')
        return msg

    def _devmsg(self):
        if self._verbosity == 2:
            info = inspect.getouterframes(inspect.currentframe())[3]
            devmsg = self._start + self._blue_CL + info[1].split('/')[-1] + ' ' + str(info[2]) \
                        + ' ' + info[3] + '()' + self._end + ' - '
        else:
            devmsg = ''
        return devmsg

    def _print(self, premsg, msg, last=True):
        """
        Print to standard error and the log file
        """
        devmsg = self._devmsg()
        _msg = premsg+devmsg+msg
        if self._verbosity != 0:
            print(_msg, file=sys.stderr)
        if self._log:
            clean_msg = self._cleancolors(_msg)
            self._log.write(clean_msg+'\n' if last else clean_msg)

    def _initialize_log_file(self, log=None):
        """
        Expects self._log is already None.
        """
        if log is None:
            return

        # Initialize the log
        self._log = open(log, 'w')

        self._log.write("------------------------------------------------------\n\n")
#        self._log.write("PypeIt was last updated {0:s}\n".format(self._last_updated))
        self._log.write("This log was generated with version {0:s} of PypeIt\n\n".format(
                                                                                    self._version))
        self._log.write("You are using scipy version={:s}\n".format(scipy.__version__))
        self._log.write("You are using numpy version={:s}\n".format(numpy.__version__))
        self._log.write("You are using astropy version={:s}\n\n".format(astropy.__version__))
        self._log.write("------------------------------------------------------\n\n")

    def reset(self, log=None, verbosity=None, colors=True):
        """
        Reinitialize the object.

        Needed so that there can be a default object for all modules,
        but also a dynamically defined log file.
        """
        # Initialize other variables
        self._verbosity = self._defverb if verbosity is None else verbosity
        self.reset_log_file(log)
        self.disablecolors()
        if colors:
            self.enablecolors()

    def reset_log_file(self, log):
        if self._log:
            self._log.close()
            self._log = None
        self._initialize_log_file(log=log)

    # Headers and usage
    # TODO: Move this to the ARMED class...
    def armedheader(self, prognm):
        """
        Get the info header for ARMED
        """
        header = '##  '
        header += self._start + self._white_GR + 'ARMED : '
        header += 'Automated Reduction and Modelling of Echelle Data v{0:s}'.format(
                        self._version) + self._end + '\n'
        header += '##  '
        header += 'Usage : '
        header += 'python %s [options] filelist'.format(prognm)
        return header

    def pypeitheader(self, prognm):
        """
        Get the info header for PypeIt
        """
        header = '##  '
        header += self._start + self._white_GR + 'PypeIt : '
        header += 'The Python Spectroscopic Data Reduction Pipeline v{0:s}'.format(
                        self._version) + self._end + '\n'
        header += '##  '
        return header

    def usage(self, prognm):
        """
        Print pypeit usage data.
        """
        # TODO: This is outdated
        stgs_arm = glob.glob(os.path.dirname(__file__)+'/data/settings/settings.arm*')
        stgs_all = glob.glob(os.path.dirname(__file__)+'/data/settings/settings.*')
        stgs_spc = list(set(stgs_arm) ^ set(stgs_all))

        armlist = os.path.basename(stgs_arm[0]).split('.')[-1]
        for i in range(1, len(stgs_arm)):
            armlist += ', ' + os.path.basename(stgs_arm[i]).split('.')[-1]
        spclist = os.path.basename(stgs_spc[0]).split('.')[-1]
        for kk,istsp in enumerate(stgs_spc):
            if (kk == 0) or ('base' in istsp) or ('py' in istsp.split('.')[-1]):
                continue
            spclist += ', ' + istsp.split('.')[-1]

        spcl = textwrap.wrap(spclist, width=60)
        descs = self.pypeitheader(prognm)

        descs += '\n##  Available pipelines include (OUTDATED):'
        descs += '\n##   ' + armlist

        descs += '\n##  Available spectrographs include (OUTDATED):'
        for ispcl in spcl:
            descs += '\n##   ' + ispcl

#        descs += '\n##  Last updated: {0:s}'.format(self._last_updated)

        return descs

    def close(self):
        '''
        Close the log file before the code exits
        '''
        close_qa(self.pypeit_file)
#        from pypeit import arqa
#        # QA HTML
#        if self.pypeit_file is not None:  # Likely testing
#            try:
#                arqa.gen_mf_html(self.pypeit_file)
#            except:  # Likely crashed very early
#                pass
#            else:
#                arqa.gen_exp_html()
        return self.reset_log_file(None)

#        # Close log
#        if self._log:
#            self._log.close()

#    def signal_handler(self, signalnum, handler):
#        """
#        Handle signals sent by the keyboard during code execution
#        """
#        if signalnum == 2:
#            self.info('Ctrl+C was pressed. Ending processes...')
#            self.close()
#            sys.exit()

    def error(self, msg, usage=False):
        """
        Print an error message
        """
        premsg = '\n'+self._start + self._white_RD + '[ERROR]   ::' + self._end + ' '
        self._print(premsg, msg)

        # Close log file
        # TODO: This no longer "closes" the QA plots
        self.close()

        # Print command line usage
        if usage:
            self.usage(None)
        #
        raise PypeItError(msg)
        sys.exit(1)

    def info(self, msg):
        """
        Print an information message
        """
        premsg = self._start + self._green_CL + '[INFO]    ::' + self._end + ' '
        self._print(premsg, msg)

    def info_update(self, msg, last=False):
        """
        Print an information message that needs to be updated
        """
        premsg = '\r' + self._start + self._green_CL + '[INFO]    ::' + self._end + ' '
        self._print(premsg, msg, last=last)

    def test(self, msg):
        """
        Print a test message
        """
        if self._verbosity == 2:
            premsg = self._start + self._white_BL + '[TEST]    ::' + self._end + ' '
            self._print(premsg, msg)

    def warn(self, msg):
        """
        Print a warning message
        """
        premsg = self._start + self._red_CL + '[WARNING] ::' + self._end + ' '
        self._print(premsg, msg)

    def bug(self, msg):
        """
        Print a bug message
        """
        premsg = self._start + self._white_BK + '[BUG]     ::' + self._end + ' '
        self._print(premsg, msg)

    def work(self, msg):
        """
        Print a work in progress message
        """
        if self._verbosity == 2:
            premsgp = self._start + self._black_CL + '[WORK IN ]::' + self._end + '\n'
            premsgs = self._start + self._yellow_CL + '[PROGRESS]::' + self._end + ' '
            self._print(premsgp+premsgs, msg)

    def prindent(self, msg):
        """
        Print an indent
        """
        premsg = '             '
        self._print(premsg, msg)

    def input(self):
        """
        Return a text string to be used to display input required from the user
        """
        premsg = self._start + self._blue_CL + '[INPUT]   ::' + self._end + ' '
        return premsg

    @staticmethod
    def newline():
        """
        Return a text string containing a newline to be used with messages
        """
        return '\n             '

    @staticmethod
    def indent():
        """
        Return a text string containing an indent to be used with messages
        """
        return '             '

    # Set the colors
    def enablecolors(self):
        """
        Enable colored output text
        """

        # Start and end coloured text
        self._start = '\x1B['
        self._end = '\x1B[' + '0m'

        # Clear Backgrounds
        self._black_CL = '1;30m'
        self._yellow_CL = '1;33m'
        self._blue_CL = '1;34m'
        self._green_CL = '1;32m'
        self._red_CL = '1;31m'

        # Coloured Backgrounds
        self._white_RD = '1;37;41m'
        self._white_GR = '1;37;42m'
        self._white_BK = '1;37;40m'
        self._white_BL = '1;37;44m'
        self._black_YL = '1;37;43m'
        self._yellow_BK = '1;33;40m'

    def disablecolors(self):
        """
        Disable colored output text
        """

        # Start and end coloured text
        self._start = ''
        self._end = ''

        # Clear Backgrounds
        self._black_CL = ''
        self._yellow_CL = ''
        self._blue_CL = ''
        self._green_CL = ''
        self._red_CL = ''

        # Coloured Backgrounds
        self._white_RD = ''
        self._white_GR = ''
        self._white_BK = ''
        self._white_BL = ''
        self._black_YL = ''
        self._yellow_BK = ''






"""
Module to run tests on datamodel.DataContainer
"""
import sys
import io
import os
import shutil
import inspect

import pytest

import numpy as np

#import pypeit

from astropy.table import Table
from astropy.io import fits

from pypeit.datamodel import DataContainer

#-----------------------------------------------------------------------
# Example derived classes
class BasicContainer(DataContainer):
    version = '1.0.0'
    datamodel = {'vec1': dict(otype=np.ndarray, atype=float, descr='Test'),
                 'meta1': dict(otype=str, decr='test'),
                 'arr1': dict(otype=np.ndarray, atype=float, descr='test')}

    def __init__(self, vec1, meta1, arr1):
        # All arguments are passed directly to the container
        # instantiation
        args, _, _, values = inspect.getargvalues(inspect.currentframe())
        super(BasicContainer, self).__init__({k: values[k] for k in args[1:]}) 

    def _bundle(self):
        # Specify the extension
        return super(BasicContainer, self)._bundle(ext='basic')


class MixedCaseContainer(DataContainer):
    version = '1.0.0'
    datamodel = {'lowercase': dict(otype=np.ndarray, atype=float, descr='Test'),
                 'UPPERCASE': dict(otype=int, decr='test'),
                 'CamelCase': dict(otype=float, decr='test')}

    def __init__(self, lowercase, UPPERCASE, CamelCase):
        # All arguments are passed directly to the container
        # instantiation
        args, _, _, values = inspect.getargvalues(inspect.currentframe())
        super(MixedCaseContainer, self).__init__({k: values[k] for k in args[1:]}) 

    def _bundle(self):
        # Specify the extension
        return super(MixedCaseContainer, self)._bundle(ext='mixedcase')


class ImageContainer(DataContainer):
    version = '1.0.0'
    datamodel = {'img1': dict(otype=np.ndarray, atype=float, descr='Test'),
                 'img1_key': dict(otype=str, descr='test'),
                 'img2': dict(otype=np.ndarray, descr='test')}

    def __init__(self, img1, img2, img1_key=None):
        # All arguments are passed directly to the container
        # instantiation
        args, _, _, values = inspect.getargvalues(inspect.currentframe())
        super(ImageContainer, self).__init__({k: values[k] for k in args[1:]}) 

    def _bundle(self):
        # img1 and key are put in the first extension, img2 in the
        # second one
        return [ {'img1_key':self['img1_key'], 'img1': self['img1']}, {'img2':self['img2']} ]


class GoodTableContainer(DataContainer):
    version = '1.0.0'
    datamodel = {'tab1': dict(otype=Table, descr='Test'),
                 'tab1len': dict(otype=int, descr='test'),
                 'tab2': dict(otype=Table, descr='test'),
                 'tab2len': dict(otype=int, descr='test')}

    def __init__(self, tab1, tab2):
        # All arguments are passed directly to the container
        # instantiation, but the list is incomplete
        args, _, _, values = inspect.getargvalues(inspect.currentframe())
        super(GoodTableContainer, self).__init__({k: values[k] for k in args[1:]}) 

    def _validate(self):
        # Complete the instantiation
        self.tab1len = len(self.tab1)
        self.tab2len = len(self.tab2)

    def _bundle(self):
        # Bundle so there's only one Table per extension
        return [{'tab1len': self.tab1len, 'tab1': self.tab1},
                {'tab2len': self.tab2len, 'tab2': self.tab2}]


class BadTableContainer(GoodTableContainer):
    def _bundle(self):
        # Use default _bundle method, which will try to put both tables
        # in the same extension.  NOTE: Can't use super here because
        # GoodTableContainer doesn't have an 'ext' argument
        return DataContainer._bundle(self, ext='bad')


class GoodMixedTypeContainer(DataContainer):
    version = '1.0.0'
    datamodel = {'tab1': dict(otype=Table, descr='Test'),
                 'tab1len': dict(otype=int, descr='test'),
                 'arr1': dict(otype=np.ndarray, descr='test'),
                 'arr1shape': dict(otype=tuple, descr='test')}

    def __init__(self, tab1, arr1):
        # All arguments are passed directly to the container
        # instantiation, but the list is incomplete
        args, _, _, values = inspect.getargvalues(inspect.currentframe())
        super(GoodMixedTypeContainer, self).__init__({k: values[k] for k in args[1:]}) 

    def _validate(self):
        # Complete the instantiation
        self.tab1len = len(self.tab1)
        # NOTE: DataContainer does allow for tuples, but beware because
        # they have to saved to the fits files by converting them to strings
        # and writing them to the fits header.  So the tuples should be
        # short!  See _bundle below, and in DataContainer.
        self.arr1shape = self.arr1.shape

    def _bundle(self):
        # Bundle so there's only one Table and one Image per extension
        return [{'tab1len': self.tab1len, 'tab1': self.tab1},
                {'arr1shape': str(self.arr1shape), 'arr1': self.arr1}]


class BadMixedTypeContainer(GoodMixedTypeContainer):
    def _bundle(self):
        # Use default _bundle method, which will try to put both tables
        # in the same extension.  NOTE: Can't use super here because
        # GoodMixedTypeContainer doesn't have an 'ext' argument
        return DataContainer._bundle(self, ext='bad')


class BadInitContainer(DataContainer):
    version = '1.0.0'
    datamodel = {'inp1': dict(otype=np.ndarray, descr='Test'),
                 'inp2': dict(otype=np.ndarray, descr='test'),
                 'out': dict(otype=np.ndarray, descr='test'),
                 'alt': dict(otype=np.ndarray, descr='test')}

    def __init__(self, inp1, inp2, func='add'):
        args, _, _, values = inspect.getargvalues(inspect.currentframe())
        super(BadInitContainer, self).__init__({k: values[k] for k in args[1:]}) 


class DubiousInitContainer(DataContainer):
    version = '1.0.0'
    datamodel = {'inp1': dict(otype=np.ndarray, descr='Test'),
                 'inp2': dict(otype=np.ndarray, descr='test'),
                 'out': dict(otype=np.ndarray, descr='test'),
                 'alt': dict(otype=np.ndarray, descr='test')}

    def __init__(self, inp1, inp2, func='add'):
        # If any of the arguments of the init method aren't actually
        # part of the datamodel, you can't use the nominal two lines
        # used in all the other examples above.  You have to be specific
        # about what gets passed to super.__init__.  See the
        # BadInitContainer example above and the test below.  WARNING:
        # I'm not sure you would ever want to do this because it can
        # lead to I/O issues; see the _validate function.
        self.func = func
        super(DubiousInitContainer, self).__init__({'inp1': inp1, 'inp2':inp2})

    def _validate(self):
        # Because func isn't part of the data model, it won't be part of
        # self if the object is instantiated from a file.  So I have to
        # add it here.  But I don't know what the value of the attribute
        # was for the original object that was written to disk.  This is
        # why you likely always want anything that's critical to setting
        # up the object to be part of the datamodel so that it gets
        # written to disk.  See the testing examples for when this will
        # go haywire.
        if not hasattr(self, 'func'):
            self.func = None
        if self.func not in [None, 'add', 'sub']:
            raise ValueError('Function must be either \'add\' or \'sub\'.')

        # This is here because I don't want to overwrite something that
        # might have been read in from an HDU, particularly given that
        # func will be None if reading from a file!
        if self.out is None:
            print('Assigning out!')
            if self.func is None:
                raise ValueError('Do not know how to construct out attribute!')
            self.out = self.inp1 + self.inp2 if self.func == 'add' else self.inp1 - self.inp2

    # I'm not going to overwrite _bundle, so that the nominal approach
    # is used.


class ComplexInitContainer(DataContainer):
    version = '1.0.0'
    datamodel = {'inp1': dict(otype=np.ndarray, descr='Test'),
                 'inp2': dict(otype=np.ndarray, descr='test'),
                 'out': dict(otype=np.ndarray, descr='test'),
                 'alt': dict(otype=np.ndarray, descr='test'),
                 'func': dict(otype=str, descr='test')}

    def __init__(self, inp1, inp2, func='add'):
        # Since func is part of the datamodel now, we can use the normal
        # two intantiation lines.
        args, _, _, values = inspect.getargvalues(inspect.currentframe())
        super(ComplexInitContainer, self).__init__({k: values[k] for k in args[1:]}) 

    def _validate(self):
        if self.func not in ['add', 'sub']:
            raise ValueError('Function must be either \'add\' or \'sub\'.')
        # This is here because I don't want to overwrite something that
        # might have been read in from an HDU, even though they should
        # nominally be the same!
        if self.out is None:
            print('Assigning out!')
            if self.func is None:
                raise ValueError('Do not know how to construct out attribute!')
            self.out = self.inp1 + self.inp2 if self.func == 'add' else self.inp1 - self.inp2

#-----------------------------------------------------------------------

def test_basic():

    # Remove file if it exists
    ofile = 'test.fits.gz'
    if os.path.isfile(ofile):
        os.remove(ofile)

    # Instantiate such that number of data table rows would be the same
    # (10)
    data = BasicContainer(np.arange(10), 'length=10', np.arange(30).reshape(10,3))

    # Instantiation and access tests
    assert list(data.keys()) == ['vec1', 'meta1', 'arr1'], 'Bad keys'
    assert np.array_equal(data.vec1, np.arange(10)), 'Bad instantiation'
    assert np.array_equal(data['vec1'], np.arange(10)), 'Bad instantiation'
    assert np.array_equal(data['arr1'], np.arange(30).reshape(10,3)), 'Bad instantiation'
    assert data.meta1 == 'length=10', 'Bad instantiation'
    assert data.version == BasicContainer.version, 'Bad version'
    # The version number cannot be changed!
    with pytest.raises(TypeError):
        data.version = '2.0'

    # Cannot add elements that aren't part of the datamodel
    with pytest.raises(KeyError):
        data['newvec'] = np.arange(10)
    with pytest.raises(AttributeError):
        data.newvec = np.arange(10)
    # Attempting to access a bad attribute raises an AttributeError
    with pytest.raises(AttributeError):
        test = data.newvec
    # Attempting to access a bad item raises a KeyError
    with pytest.raises(KeyError):
        test = data['newvec']

    # Strict types enforced
    with pytest.raises(TypeError):
        data.vec1 = 3
    with pytest.raises(TypeError):
        data.meta1 = 4.

    # Write to a file
    data.to_file(ofile)

    # Test written data against input
    with fits.open(ofile) as hdu:
        assert len(hdu) == 2, 'Should be two extensions'
        assert hdu[1].name == 'BASIC', 'Incorrect extension Name'
        assert len(hdu[1].data) == 10, 'Incorrect number of rows'
        assert hdu[1].columns.names == ['vec1', 'arr1'], 'Incorrect column names'
        assert 'meta1' in hdu[1].header, 'Missing header keyword'
        assert hdu[1].header['meta1'] == 'length=10', 'Incorrect header keyword'
   
    # Remove it
    os.remove(ofile)

    # Rows of arrays mismatch, so data is stuffed into a single table
    # row
    data = BasicContainer(np.arange(10), 'length=1', np.arange(30).reshape(3,10))
    data.to_file(ofile)

    with fits.open(ofile) as hdu:
        assert len(hdu) == 2, 'Should be two extensions'
        assert hdu[1].name == 'BASIC', 'Incorrect extension Name'
        assert len(hdu[1].data) == 1, 'Incorrect number of rows'
        assert hdu[1].columns.names == ['vec1', 'arr1'], 'Incorrect column names'
        assert 'meta1' in hdu[1].header, 'Missing header keyword'
        assert hdu[1].header['meta1'] == 'length=1', 'Incorrect header keyword'

    # Test file input
    _data = BasicContainer.from_file(ofile)
    assert np.array_equal(data.vec1, _data.vec1), 'Bad read'
    assert np.array_equal(data['arr1'], _data['arr1']), 'Bad read'
    assert data.meta1 == _data['meta1'], 'Bad read'
    
    # Clean up
    os.remove(ofile)


def test_case():
    # Remove file if it exists
    ofile = 'test.fits.gz'
    if os.path.isfile(ofile):
        os.remove(ofile)

    # Instantiate
    data = MixedCaseContainer(np.arange(10), 5, 8.2)

    # Attributes and items are case-sensitive
    with pytest.raises(AttributeError):
        data.uppercase = 3
    with pytest.raises(KeyError):
        data['LOWERCASE'] = 3
    with pytest.raises(AttributeError):
        data.cAMELcASE = 3

    # HDU items are not
    data.to_file(ofile)
    with fits.open(ofile) as hdu:
        assert hdu['mixedcase'].header['UpPeRcAsE'] == 5, 'Bad header access'
        assert hdu['mixedcase'].header['CaMeLcAsE'] == 8.2, 'Bad header access'

    # And everything is as it should be when you read it back in
    _data = MixedCaseContainer.from_file(ofile)
    assert _data.UPPERCASE == 5, 'Bad file read'

    # Clean up
    os.remove(ofile)


def test_image():

    img = ImageContainer(np.arange(100).reshape(10,10), np.arange(25).reshape(5,5), img1_key='test')
    hdu = img.to_hdu(add_primary=True)
    _img = ImageContainer.from_hdu(hdu)

    assert img.__dict__.keys() == _img.__dict__.keys(), 'Keys are different'
    assert len(hdu) == 3, 'Should be 3 extensions.'

    # No keyword defined
    img = ImageContainer(np.arange(100).reshape(10,10), np.arange(25).reshape(5,5))
    assert img.img1_key is None, 'Bad instantiation'
    hdu = img.to_hdu(add_primary=True)
    _img = ImageContainer.from_hdu(hdu)
    assert _img.img1_key is None, 'Bad read'


def test_table():

    x = np.arange(10)
    y = np.arange(10)+5
    z = np.arange(30).reshape(10,3)

    a = np.arange(15).reshape(3,5)
    b = np.zeros(3)
    c = np.full((3,3,3), -1)

    tab1 = Table(data=({'x':x,'y':y,'z':z}), meta={'test':'this'})
    tab2 = Table(data=({'a':a,'b':b,'c':c}), meta={'test':'that'})

    # Check the instantiation
    data = GoodTableContainer(tab1, tab2)
    assert data.tab1.keys() == ['x', 'y', 'z'], 'Bad table save'
    assert data.tab1.meta['test'] == 'this', 'Bad table meta save'

    # Write it to an HDUList
    hdu = data.to_hdu(add_primary=True)
    assert len(hdu) == 3, 'Should be 3 extensions'
    assert [h.name for h in hdu] == ['PRIMARY', 'TAB1', 'TAB2'], 'Bad extension names'
    assert len(hdu['TAB1'].data) == data.tab1len, 'Bad table length'
    assert len(hdu['TAB2'].data) == data.tab2len, 'Bad table length'
    assert hdu['TAB2'].data['c'].shape == (3,3,3), 'Bad column shape'
    assert hdu['TAB2'].header['tab2len'] == 3, 'Bad header data'
    assert hdu['TAB1'].header['TAB1LEN'] == 10, 'Bad header data'

    # Table metadata is also written to the header, which can be
    # accessed with case-insensitive keys
    assert hdu['TAB1'].header['TEST'] == 'this', 'Bad table meta data save'
    assert hdu['TAB1'].header['TesT'] == 'this', 'Bad table meta data save'
    # But the keyword case gets mangled when you read it back in.  This
    # has to do with the Table.read method and I'm not sure there's
    # anything we can do about it without the help of the astropy folks
    _data = GoodTableContainer.from_hdu(hdu)
    with pytest.raises(KeyError):
        _data.tab1.meta['test'] == 'this'
    assert _data.tab1.meta['TEST'] == 'this', 'meta data keywords are re-read in all caps'

    # The BadTableContainer will instantiate fine
    data = BadTableContainer(tab1, tab2)
    assert data.tab1.keys() == ['x', 'y', 'z'], 'Bad table save'
    assert data.tab1.meta['test'] == 'this', 'Bad table meta save'
    # But then it will barf when you try to reformat/write the data
    # because you can't write more than one table to a single HDU
    with pytest.raises(ValueError):
        hdu = data.to_hdu()

def test_mixed():

    x = np.arange(10)
    y = np.arange(10)+5
    z = np.arange(30).reshape(10,3)

    arr1 = np.full((3,3,3), -1)

    tab1 = Table(data=({'x':x,'y':y,'z':z}), meta={'test':'this'})

    # Check the instantiation
    data = GoodMixedTypeContainer(tab1, arr1)
    assert data.tab1.keys() == ['x', 'y', 'z'], 'Bad table save'
    assert data.tab1.meta['test'] == 'this', 'Bad table meta save'
    assert data.arr1shape == (3,3,3), 'Bad shape'

    # Write it to an HDUList
    hdu = data.to_hdu(add_primary=True)
    assert len(hdu) == 3, 'Should be 3 extensions'
    assert [h.name for h in hdu] == ['PRIMARY', 'TAB1', 'ARR1'], 'Bad extension names'
    assert len(hdu['TAB1'].data) == data.tab1len, 'Bad table length'
    assert hdu['ARR1'].shape == data.arr1shape, 'Bad array shape'
    assert hdu['TAB1'].header['TAB1LEN'] == 10, 'Bad header data'
    # Tuples get converted to strings
    assert hdu['ARR1'].header['ARR1SHAPE'] == str((3,3,3)), 'Bad array shape'

    # And they get converted back when they're read in.
    _data = GoodMixedTypeContainer.from_hdu(hdu)
    assert _data.arr1shape == (3,3,3), 'Bad array shape'

    # The BadMixedTypeContainer will instantiate fine
    data = BadMixedTypeContainer(tab1, arr1)
    assert data.tab1.keys() == ['x', 'y', 'z'], 'Bad table save'
    assert data.tab1.meta['test'] == 'this', 'Bad table meta save'
    assert data.arr1shape == (3,3,3), 'Bad shape'
    # But then it will barf when you try to reformat/write the data
    # because you can't write both a Table and an array to a single HDU
    with pytest.raises(ValueError):
        hdu = data.to_hdu()

def test_init():

    x = np.arange(10)
    y = np.arange(10)+5

    # Instantiation of the BadInitContainer should fail because the init
    # arguments all need to be part of the datamodel for it to work.
    with pytest.raises(AttributeError):
        data = BadInitContainer(x,y)

    # This instantiation is fine because DubiousInitContainer handles
    # the fact that some of the arguments to __init__ are not part of
    # the datamodel.
    data = DubiousInitContainer(x,y)
    assert np.array_equal(data.out, data.inp1+data.inp2), 'Bad init'
    # One component of the data model wasn't instantiated, so it will be
    # None
    assert data.alt is None, 'Bad empty datamodel component'

    # Use the other option
    data = DubiousInitContainer(x,y,func='sub')
    assert np.array_equal(data.out, np.full(10,-5)), 'Bad init'

    # Note that this instantiation will set ``out`` because it's None on
    # input.
    stdout = sys.stdout
    sys.stdout = io.StringIO()
    data = DubiousInitContainer(x,y)
    output = sys.stdout.getvalue()
    sys.stdout = stdout
    assert output.strip() == 'Assigning out!', 'Bad instantiation'

    # But if we write it out and read it back in again, the
    # initialization of out will get skipped because it's been read in
    # from the output file.
    hdu = data.to_hdu(add_primary=True)
    stdout = sys.stdout
    sys.stdout = io.StringIO()
    _data = DubiousInitContainer.from_hdu(hdu)
    output = sys.stdout.getvalue()
    sys.stdout = stdout
    assert output.strip() != 'Assigning out!', 'Bad instantiation'

    # The nuance here is that if you instantiate it, change its value
    # away from what it's internals suggest it should be, and write it,
    # it won't be what you expect when you read it back in.  This is a
    # trivial example, but it just points out that while we're strict
    # about the datamodel in these containers, we're not strict about
    # the consistency of the datamodel with respect to the internal
    # operations within the object...
    data = DubiousInitContainer(x,y)
    data.out *= 100
    data.alt = np.arange(20)
    hdu = data.to_hdu(add_primary=True)
    assert not np.array_equal(DubiousInitContainer.from_hdu(hdu).out,
                              DubiousInitContainer(x,y).out), 'Bad read'

    # The reason the DubiousInitContainer is called that way is because
    # it has attributes that cannot be reinstantiated from what's
    # written to disk, meaning these DataContainers aren't really fully
    # formed objects unless all of its relevant attributes are
    # components of the data model.
    _data = DubiousInitContainer.from_hdu(hdu)
    with pytest.raises(AssertionError):
        assert _data.func == DubiousInitContainer(x,y).func

    # This is solved by adding func to the datamodel
    data = ComplexInitContainer(x,y)
    _data = ComplexInitContainer.from_hdu(data.to_hdu(add_primary=True))
    assert data.func == _data.func, 'Bad read'




#!/usr/bin/env python
#
# See top-level LICENSE file for Copyright information
#
# -*- coding: utf-8 -*-
"""
This script enables the viewing of a FITS file
"""
import argparse

def parser(options=None):

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('file', type = str, default = None, help = 'FITS file')
    parser.add_argument('spectrograph', type=str, help='Name of the spectrograph')
    parser.add_argument("--list", default=False, help="List the extensions only?", action="store_true")
    parser.add_argument('--exten', type=int, default = 0, help="FITS extension")
    parser.add_argument('--det', type=int, default=1, help="Detector number")

    if options is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(options)
    return args

def main(args):

    import subprocess

    from astropy.io import fits

    from pypeit import msgs
    from pypeit.spectrographs import keck_lris
    from pypeit.spectrographs import keck_deimos
    from pypeit.spectrographs import gemini_gmos
    from pypeit import msgs
    from pypeit import ginga

    # List only?
    if args.list:
        hdu = fits.open(args.file)
        print(hdu.info())
        return


    # Setup for PYPIT imports
    msgs.reset(verbosity=2)

    # RAW_LRIS??
    if 'keck_lris' in args.spectrograph:
        #
        gen_lris = keck_lris.KeckLRISRSpectrograph()  # Using LRISr, but this will work for LRISb too
        img, _, _, _, _ = gen_lris.get_rawimage(args.file,  None)
    # RAW_DEIMOS??
    elif args.spectrograph == 'keck_deimos':
        #
        gen_deimos = keck_deimos.KeckDEIMOSSpectrograph()
        img, _, _, _, _ = gen_deimos.get_rawimage(args.file, None)
    # RAW_GEMINI??
    elif 'gemini_gmos' in args.spectrograph:
        # TODO this routine should show the whole mosaic if no detector number is passed in!
        # Need to figure out the number of amps
        gen_gmos = gemini_gmos.GeminiGMOSSpectrograph()
        img, _, _, _, _ = gen_gmos.get_rawimage(args.file, args.det)
    else:
        hdu = fits.open(args.file)
        img = hdu[args.exten].data
        # Write

    ginga.show_image(img)



"""
Module to create models of arc lines.
"""
import astropy
import re
import scipy

import numpy as np
import matplotlib.pyplot as plt

from pkg_resources import resource_filename

from astropy.io import fits
from astropy.convolution import convolve, Gaussian1DKernel
from astropy.table import Table
from astropy import units

from pypeit import msgs
from pypeit.core import arc
from pypeit import utils
from pypeit.core.wave import airtovac

from IPython import embed

def blackbody(wavelength, T_BB=250., debug=False):
    """ Given wavelength [in microns] and Temperature in Kelvin
    it returns the black body emission.

    Parameters
    ----------
    wavelength : np.array
        wavelength vector in microns
    T_BB : float 
        black body temperature in Kelvin. Default is set to:
        T_BB = 250.

    Returns
    -------
    blackbody : np.array
        spectral radiance of the black body in cgs units:
        B_lambda = 2.*h*c^2/lambda^5.*(1./(exp(h*c/(lambda*k_b*T_BB))-1.)
    blackbody_counts : np.array
        Same as above but in flux density
    """

    # Define constants in cgs
    PLANCK  = astropy.constants.h.cgs.value    # erg*s
    C_LIGHT = astropy.constants.c.cgs.value    # cm/s
    K_BOLTZ = astropy.constants.k_B.cgs.value  # erg/K
    RADIAN_PER_ARCSEC = 1./3600.*np.pi/180.

    msgs.info("Creating BB spectrum at T={}K".format(T_BB))
    lam = wavelength / 1e4 # convert wave in cm.
    blackbody_pol = 2.*PLANCK*np.power(C_LIGHT,2) / np.power(lam,5)
    blackbody_exp = np.exp(PLANCK*C_LIGHT/(lam*K_BOLTZ*T_BB)) - 1.
    blackbody = blackbody_pol / blackbody_exp
    blackbody_counts = blackbody / (PLANCK * C_LIGHT / lam) * 1e-4 \
                 * np.power(RADIAN_PER_ARCSEC, 2.)

    if debug:
        utils.pyplot_rcparams()
        msgs.info("Plot of the blackbody spectrum.")
        plt.figure()
        plt.plot(wavelength, blackbody,
                 color='navy', linestyle='-', alpha=0.8,
                 label=r'T_BB={}'.format(T_BB))
        plt.legend()
        plt.xlabel(r"Wavelength [micron]")
        plt.ylabel(r"Spectral Radiance")
        plt.title(r"Planck's law")
        msgs.info("Close the Figure to continue.")
        plt.show(block=True)
        plt.close()
        utils.pyplot_rcparams_default()

    return blackbody, blackbody_counts


def addlines2spec(wavelength, wl_line, fl_line, resolution,
                  scale_spec=1., debug=False):
    """ Create a spectrum with a set of (gaussian) emission lines.
    
    Parameters
    ----------
    wavelength : np.array
        wavelength vector of the input spectrum
    wl_line, fl_line : np.arrays
        wavelength and flux of each individual line
    resolution : np.float
        resolution of the spectrograph. In other words, the lines
        will have a FWHM equal to:
        fwhm_line = wl_line / resolution
    scale_spec : np.float
        rescale all the  normalization of the final spectrum.
        Default scale_spec=1.
    debug : boolean
        If True will show debug plots

    Returns
    -------
    line_spec : np.array
        Spectrum with lines

    """
    line_spec = np.zeros_like(wavelength)
    wl_line_min, wl_line_max = np.min(wavelength), np.max(wavelength)
    good_lines = (wl_line>wl_line_min) & (wl_line<wl_line_max)
    wl_line_good = wl_line[good_lines]
    fl_line_good = fl_line[good_lines]

    # define sigma of the gaussians
    sigma = wl_line_good / resolution / 2.355

    msgs.info("Creating line spectrum")
    for ii in np.arange(len(wl_line_good)):
        line_spec += scale_spec*fl_line_good[ii]*\
                     np.exp(-np.power((wl_line_good[ii]-wavelength),2.)/(2.*np.power(sigma[ii],2.)))

    if debug:
        utils.pyplot_rcparams()
        msgs.info("Plot of the line spectrum.")
        plt.figure()
        plt.plot(wavelength, line_spec,
                 color='navy', linestyle='-', alpha=0.8,
                 label=r'Spectrum with lines included')
        plt.legend()
        plt.xlabel(r'Wavelength')
        plt.ylabel(r'Flux')
        msgs.info("Close the Figure to continue.")
        plt.show(block=True)
        plt.close()
        utils.pyplot_rcparams_default()

    return line_spec


def oh_lines():
    """ Reads in the Rousselot (2000) OH line list"

    Returns
    -------
    wavelength, amplitude : np.arrays
        Wavelength [in microns] and amplitude of the OH lines.
    """

    msgs.info("Reading in the Rousselot (2000) OH line list")
    skisim_dir = resource_filename('pypeit', 'data/skisim/')
    oh = np.loadtxt(skisim_dir+"rousselot2000.dat", usecols=(0, 1))
    return oh[:,0]/10000., oh[:,1] # wave converted to microns


def transparency(wavelength, debug=False):
    """ Interpolate the atmospheric transmission model in the IR over
    a given wavelength (in microns) range.

    Parameters
    ----------
    wavelength : np.array
        wavelength vector in microns
    debug : boolean
        If True will show debug plots

    Returns
    -------
    transparency : np.array
        Transmission of the sky over the considered wavelength rage.
        1. means fully transparent and 0. fully opaque
    """

    msgs.info("Reading in the atmospheric transmission model")
    skisim_dir = resource_filename('pypeit', 'data/skisim/')
    transparency = np.loadtxt(skisim_dir+'atm_transmission_secz1.5_1.6mm.dat')
    wave_mod = transparency[:,0]
    tran_mod = transparency[:,1]

    # Limit model between 0.8 and np.max(wavelength) microns
    filt_wave_mod = (wave_mod>0.8) & (wave_mod<np.max(wavelength))
    wave_mod = wave_mod[filt_wave_mod]
    tran_mod = tran_mod[filt_wave_mod]

    # Interpolate over input wavelengths
    interp_tran = scipy.interpolate.interp1d(wave_mod, tran_mod,
                                             kind='cubic',
                                             fill_value='extrapolate')
    transmission = interp_tran(wavelength)
    transmission[wavelength<0.9] = 1.

    # Clean for spourious values due to interpolation
    transmission[transmission<0.] = 0.
    transmission[transmission>1.] = 1.

    if debug:
        utils.pyplot_rcparams()
        msgs.info("Plot of the sky transmission template")
        plt.figure()
        plt.plot(wave_mod, tran_mod,
                 color='navy', linestyle='-', alpha=0.8,
                 label=r'Original')
        plt.plot(wavelength, transmission, 
                 color='crimson', linestyle='-', alpha=0.8,
                 label=r'Resampled')
        plt.legend()
        plt.xlabel(r'Wavelength [microns]')
        plt.ylabel(r'Transmission')
        plt.title(r' IR Transmission Spectra ')
        msgs.info("Close the Figure to continue.")
        plt.show(block=True)
        plt.close()
        utils.pyplot_rcparams_default()

    # Returns
    return transmission


def h2o_lines():
    """ Reads in the H2O atmospheric spectrum"

    Returns
    -------
    wavelength, flux : np.arrays
        Wavelength [in microns] and flux of the H2O atmospheric 
        spectrum.
    """

    msgs.info("Reading in the water atmsopheric spectrum")
    skisim_dir = resource_filename('pypeit', 'data/skisim/')
    h2o = np.loadtxt(skisim_dir+"HITRAN.txt", usecols=(0, 1))
    h2o_wv = 1./ h2o[:,0] * 1e4 # microns
    h2o_rad = h2o[:,1] * 5e11 # added to match XIDL

    return h2o_wv, h2o_rad


def thar_lines():
    """ Reads in the H2O atmospheric spectrum"
    Detailed information are here: http://astronomy.swin.edu.au/~mmurphy/thar/index.html

    Returns
    -------
    wavelength, flux : np.arrays
        Wavelength [in angstrom] and flux of the ThAr lamp 
        spectrum.
    """

    msgs.info("Reading in the ThAr spectrum")
    arclines_dir = resource_filename('pypeit', 'data/arc_lines/')
    thar = fits.open(arclines_dir+'thar_spec_MM201006.fits')

    # create pixel array
    thar_pix = np.arange(thar[0].header['CRPIX1'],len(thar[0].data[0,:])+1)
    # convert pixels to wavelength in Angstrom
    thar_wv = thar[0].header['UP_WLSRT']*10**((thar_pix-thar[0].header['CRPIX1'])*thar[0].header['CD1_1'])
    # read in spectrum
    thar_spec = thar[0].data[0,:]

    return thar_wv, thar_spec


def nearIR_modelsky(resolution, waveminmax=(0.8,2.6), dlam=40.0,
                    flgd=True, nirsky_outfile=None, T_BB=250.,
                    SCL_BB=1., SCL_OH=1., SCL_H2O=10.,
                    WAVE_WATER=2.3, debug=False):
    """ Generate a model sky in the near-IR. This includes a continuum model
    to match to gemini broadband level, a black body at T_BB, OH lines, and 
    H2O lines (but only at lambda>WAVE_WATER). Everythins is smoothed at the
    given resolution.

    Parameters
    ----------
    resolution : np.float
        resolution of the spectrograph. The OH and H2O lines will have a 
        FWHM equal to:
        fwhm_line = wl_line / resolution
    waveminmax : tuple
        wavelength range in microns to be covered by the model.
        Default is: (0.8, 2.6)
    dlam : 
        bin to be used to create the wavelength grid of the model.
        If flgd='True' it is a bin in velocity (km/s). If flgd='False'
        it is a bin in linear space (microns).
        Default is: 40.0 (with flgd='True')
    flgd : boolean
        if flgd='True' (default) wavelengths are created with 
        equal steps in log space. If 'False', wavelengths will be
        created wit equal steps in linear space.
    nirsky_outfile : str
        name of the fits file where the model sky spectrum will be stored.
        default is 'None' (i.e., no file will be written).
    T_BB : float 
        black body temperature in Kelvin. Default is set to:
        T_BB = 250.
    SCL_BB : float
        scale factor for modelling the sky black body emssion.
        Default: SCL_BB=1.
    SCL_OH : float
        scale factor for modelling the OH emssion.
        Default: SCL_OH=1.
    SCL_H2O : float
        scale factor for modelling the H2O emssion.
        Default: SCL_H2O=10.
    WAVE_WATER : float
        wavelength (in microns) at which the H2O are inclued.
        Default: WAVE_WATER = 2.3
    debug : boolean
        If True will show debug plots

    Returns
    -------
    wave, sky_model : np.arrays
        wavelength (in Ang.) and flux of the final model of the sky.
    """

    # Create the wavelength array:
    wv_min = waveminmax[0]
    wv_max = waveminmax[1]
    if flgd :
        msgs.info("Creating wavelength vector in velocity space.")
        velpix = dlam # km/s
        loglam = np.log10(1.0 + velpix/299792.458)
        wave = np.power(10.,np.arange(np.log10(wv_min), np.log10(wv_max), loglam))
    else :
        msgs.info("Creating wavelength vector in linear space.")
        wave = np.arange(wv_min, wv_max, dlam)

    # Calculate transparency
    # trans = transparency(wave, debug=False)

    # Empirical match to gemini broadband continuum level
    logy = - 0.55 - 0.55 * (wave-1.0)
    y = np.power(10.,logy)

    msgs.info("Add in a blackbody for the atmosphere.")
    bb, bb_counts = blackbody(wave, T_BB=T_BB, debug=debug)
    bb_counts = bb_counts

    msgs.info("Add in OH lines")
    oh_wv, oh_fx = oh_lines()
    # produces better wavelength solutions with 1.0 threshold
    msgs.info("Selecting stronger OH lines")
    filt_oh = oh_fx > 1.
    oh_wv, oh_fx = oh_wv[filt_oh], oh_fx[filt_oh]
    # scale_spec was added to match the XIDL code
    ohspec = addlines2spec(wave, oh_wv, oh_fx, resolution=resolution,
                           scale_spec=((resolution/1000.)/40.),
                           debug=debug)

    if wv_max > WAVE_WATER :
        msgs.info("Add in H2O lines")
        h2o_wv, h2o_rad = h2o_lines()
        filt_h2o = (h2o_wv>wv_min-0.1) & (h2o_wv<wv_max+0.1)
        h2o_wv  = h2o_wv[filt_h2o]
        h2o_rad = h2o_rad[filt_h2o]
        # calculate sigma at the mean wavelenght of the H2O spectrum
        filt_h2o_med = h2o_wv>WAVE_WATER
        mn_wv = np.mean(h2o_wv[filt_h2o_med])
        # Convolve to the instrument resolution. This is only
        # approximate.
        smooth_fx, dwv, h2o_dwv = conv2res(h2o_wv, h2o_rad,
                                           resolution,
                                           central_wl = mn_wv,
                                           debug=debug)
        # Interpolate over input wavelengths
        interp_h2o = scipy.interpolate.interp1d(h2o_wv, smooth_fx,
                                                kind='cubic', 
                                                fill_value='extrapolate')
        h2ospec = interp_h2o(wave)
        # Zero out below WAVE_WATER microns (reconsider)
        h2ospec[wave<WAVE_WATER] = 0.
        h2ospec[wave>np.max(h2o_wv)] = 0.
    else:
        h2ospec = np.zeros(len(wave),dtype='float')

    sky_model = y+bb_counts*SCL_BB+ohspec*SCL_OH+h2ospec*SCL_H2O

    if nirsky_outfile is not None:
        msgs.info("Saving the sky model in: {}".format(nirsky_outfile))
        hdu = fits.PrimaryHDU(np.array(sky_model))
        header = hdu.header
        if flgd :
            header['CRVAL1'] = np.log10(wv_min)
            header['CDELT1'] = loglam
            header['DC-FLAG'] = 1
        else :
            header['CRVAL1'] = wv_min
            header['CDELT1'] = dlam
            header['DC-FLAG'] = 0
        hdu.writeto(nirsky_outfile, overwrite = True)

    if debug:
        utils.pyplot_rcparams()
        msgs.info("Plot of the sky emission at R={}".format(resolution))
        plt.figure()
        plt.plot(wave, sky_model,
                 color='black', linestyle='-', alpha=0.8,
                 label=r'Sky Model')
        plt.plot(wave, y,
                 color='darkorange', linestyle='-', alpha=0.6,
                 label=r'Continuum')
        plt.plot(wave, bb_counts*SCL_BB, 
                 color='green', linestyle='-', alpha=0.6,
                 label=r'Black Body at T={}K'.format(T_BB))
        plt.plot(wave, ohspec*SCL_OH,
                 color='darkviolet', linestyle='-', alpha=0.6,
                 label=r'OH')
        plt.plot(wave, h2ospec*SCL_H2O,
                 color='dodgerblue', linestyle='-', alpha=0.6,
                 label=r'H2O')
        plt.legend()
        plt.xlabel(r'Wavelength [microns]')
        plt.ylabel(r'Emission')
        plt.title(r'Sky Emission Spectrum at R={}'.format(resolution))
        msgs.info("Close the Figure to continue.")
        plt.show(block=True)
        plt.close()
        utils.pyplot_rcparams_default()

    return np.array(wave*10000.), np.array(sky_model)


def optical_modelThAr(resolution, waveminmax=(3000.,10500.), dlam=40.0,
                      flgd=True, thar_outfile=None, debug=False):
    """ Generate a model of a ThAr lamp in the uvb/optical. This is based on the
    Murphy et al. ThAr spectrum. Detailed information are here:
    http://astronomy.swin.edu.au/~mmurphy/thar/index.html
    Everythins is smoothed at the given resolution.

    Parameters
    ----------
    resolution : np.float
        resolution of the spectrograph. The ThAr lines will have a 
        FWHM equal to:
        fwhm_line = wl_line / resolution
    waveminmax : tuple
        wavelength range in angstrom to be covered by the model.
        Default is: (3000.,10500.)
    dlam : 
        bin to be used to create the wavelength grid of the model.
        If flgd='True' it is a bin in velocity (km/s). If flgd='False'
        it is a bin in linear space (microns).
        Default is: 40.0 (with flgd='True')
    flgd : boolean
        if flgd='True' (default) wavelengths are created with 
        equal steps in log space. If 'False', wavelengths will be
        created wit equal steps in linear space.
    thar_outfile : str
        name of the fits file where the model sky spectrum will be stored.
        default is 'None' (i.e., no file will be written).
    debug : boolean
        If True will show debug plots

    Returns
    -------
    wave, thar_model : np.arrays
        wavelength (in Ang.) and flux of the final model of the ThAr lamp emission.
    """

    # Create the wavelength array:
    wv_min = waveminmax[0]
    wv_max = waveminmax[1]
    if flgd :
        msgs.info("Creating wavelength vector in velocity space.")
        velpix = dlam # km/s
        loglam = np.log10(1.0 + velpix/299792.458)
        wave = np.power(10.,np.arange(np.log10(wv_min), np.log10(wv_max), loglam))
    else :
        msgs.info("Creating wavelength vector in linear space.")
        wave = np.arange(wv_min, wv_max, dlam)

    msgs.info("Add in ThAr lines")
    th_wv, th_fx = thar_lines()

    # select spectral region
    filt_wl = (th_wv>=wv_min) & (th_wv<=wv_max)
    # calculate sigma at the mean wavelenght of the ThAr spectrum
    mn_wv = np.mean(th_wv[filt_wl])
    # Convolve to the instrument resolution. This is only
    # approximate.
    smooth_fx, dwv, thar_dwv = conv2res(th_wv, th_fx,
                                        resolution,
                                        central_wl = mn_wv,
                                        debug=debug)
    # Interpolate over input wavelengths
    interp_thar = scipy.interpolate.interp1d(th_wv, smooth_fx,
                                             kind='cubic',
                                             fill_value='extrapolate')
    thar_spec = interp_thar(wave)

    # remove negative artifacts
    thar_spec[thar_spec<0.] = 0.
    # Remove regions of the spectrum outside the wavelength covered by the ThAr model
    if wv_min<np.min(th_wv):
        msgs.warn("Model of the ThAr spectrum outside the template coverage.")
        thar_spec[wave<np.min(th_wv)] = 0.
    if wv_max<np.max(th_wv):
        msgs.warn("Model of the ThAr spectrum outside the template coverage.")
        thar_spec[wave>np.max(th_wv)] = 0.

    if thar_outfile is not None:
        msgs.info("Saving the ThAr model in: {}".format(thar_outfile))
        hdu = fits.PrimaryHDU(np.array(thar_spec))
        header = hdu.header
        if flgd :
            header['CRVAL1'] = np.log10(wv_min)
            header['CDELT1'] = loglam
            header['DC-FLAG'] = 1
        else :
            header['CRVAL1'] = wv_min
            header['CDELT1'] = dlam
            header['DC-FLAG'] = 0
        hdu.writeto(thar_outfile, overwrite = True)

    if debug:
        utils.pyplot_rcparams()
        msgs.info("Plot of the Murphy et al. template at R={}".format(resolution))
        plt.figure()
        plt.plot(th_wv, th_fx,
                 color='navy', linestyle='-', alpha=0.3,
                 label=r'Original')
        plt.plot(th_wv, smooth_fx, 
                 color='crimson', linestyle='-', alpha=0.6,
                 label=r'Convolved at R={}'.format(resolution))
        plt.plot(wave, thar_spec, 
                 color='maroon', linestyle='-', alpha=1.0,
                 label=r'Convolved at R={} and resampled'.format(resolution))
        plt.legend()
        plt.xlabel(r'Wavelength [Ang.]')
        plt.ylabel(r'Emission')
        plt.title(r'Murphy et al. ThAr spectrum at R={}'.format(resolution))
        msgs.info("Close the Figure to continue.")
        plt.show(block=True)
        plt.close()
        utils.pyplot_rcparams_default()

    return np.array(wave), np.array(thar_spec)


def conv2res(wavelength, flux, resolution, central_wl='midpt',
             debug=False):
    """Convolve an imput spectrum to a specific resolution. This is only
    approximate. It takes a fix FWHM for the entire spectrum given by:
    fwhm = wl_cent / resolution

    Parameters
    ----------
    wavelength : np.array
        wavelength
    flux : np.array
        flux
    resolution : np.float
        resolution of the spectrograph
    central_wl 
        if 'midpt' the central pixel of wavelength is used, otherwise
        the central_wl will be used.
    debug : boolean
        If True will show debug plots

    Returns
    -------
    flux_convolved :np.array
        Resulting flux after convolution
    px_sigma : float
        Size of the sigma in pixels at central_wl
    px_bin : float
        Size of one pixel at central_wl
    """

    if central_wl is 'midpt':
        wl_cent = np.median(wavelength)
    else:
        wl_cent = np.float(central_wl)
    wl_sigma =  wl_cent / resolution / 2.355
    wl_bin = np.abs((wavelength - np.roll(wavelength,1))[np.where( np.abs(wavelength-wl_cent) == np.min(np.abs(wavelength-wl_cent)) )])
    msgs.info("The binning of the wavelength array at {} is: {}".format(wl_cent, wl_bin[0]))
    px_bin = wl_bin[0]
    px_sigma = wl_sigma / px_bin

    msgs.info("Covolving with a Gaussian kernel with sigma = {} pixels".format(px_sigma))
    gauss_kernel = Gaussian1DKernel(px_sigma)

    flux_convolved = convolve(flux, gauss_kernel)

    if debug:
        utils.pyplot_rcparams()
        msgs.info("Spectrum Convolved at R = {}".format(resolution))
        plt.figure()
        plt.plot(wavelength, flux,
                 color='navy', linestyle='-', alpha=0.8,
                 label=r'Original')
        plt.plot(wavelength, flux_convolved, 
                 color='crimson', linestyle='-', alpha=0.8,
                 label=r'Convolved')
        plt.legend()
        plt.xlabel(r'Wavelength')
        plt.ylabel(r'Flux')
        plt.title(r'Spectrum Convolved at R = {}'.format(resolution))
        msgs.info("Close the Figure to continue.")
        plt.show(block=True)
        plt.close()
        utils.pyplot_rcparams_default()

    return flux_convolved, px_sigma, px_bin


def iraf_datareader(database_dir, id_file):
    """Reads in a line identification database created with IRAF
    identify. These are usually locate in a directory called 'database'.
    This read pixel location and wavelength of the lines that
    have been id with IRAF. Note that the first pixel in IRAF
    is '1', while is '0' in python. The pixel location is thus
    shifted of one pixel while reading the database.

    Parameters
    ----------
    database_dir : string
        directory where the id files are located.
    id_file : string
        filename that is going to be read.

    Returns
    -------
    pixel, line_id : np.arrays
        Position of the line in pixel and ID of the line. 
        For IRAF output, these are usually in Ang.
    """

    lines_database = []

    # Open file for reading of text data.
    with open (database_dir+id_file, 'rt') as id_file_iraf:
        for line in id_file_iraf:
            lines_database.append(line.split())
            feat_line = re.search(r'features\t(\d+)', line)
            if feat_line is not None:
                N_lines = np.int(feat_line.group(1))

    msgs.info("The number of IDs in the IRAF database {} is {}".format(id_file, N_lines))

    pixel = np.zeros(N_lines)
    line_id = np.zeros(N_lines)
    for iii in range(0,N_lines):
        pixel[iii] = lines_database[10:N_lines+10][iii][0]
        line_id[iii] = lines_database[10:N_lines+10][iii][2]

    # Moving from IRAF 1-based to Python 0-based convention.
    pixel = pixel - 1.

    return pixel, line_id


def create_linelist(wavelength, spec, fwhm, sigdetec=2.,
                    cont_samp=10., line_name=None, file_root_name=None,
                    iraf_frmt=False, debug=False, vacuum=True):
    """ Create list of lines detected in a spectrum in a PypeIt
    compatible format. The name of the output file is
    file_root_name+'_lines.dat'.

    Parameters
    ----------
    wavelength : np.array
        wavelength
    spec : np.array
        spectrum
    fwhm : float
        fwhm in pixels used for filtering out arc lines that are too
        wide and not considered in fits. Parameter of arc.detect_lines().
    sigdetec : float
        sigma threshold above fluctuations for line detection. Parameter
        of arc.detect_lines(). Default = 2.
    cont_samp : float
        the number of samples across the spectrum used for continuum
        subtraction. Parameter of arc.detect_lines().  Default = 10.
    line_name : str
        name of the lines to listed in the file.
    file_root_name : str
        name of the file where the identified lines will be stored.
        The code automatically add '_lines.dat' at the end of the
        root name.
    iraf_frmt : bool
        if True, the file is written in the IRAF format (i.e. wavelength,
        ion name, amplitude).
    vacuum (bool):
        If True, write the wavelengths in vacuum
    """

    msgs.info("Searching for peaks {} sigma above background".format(sigdetec))
    tampl_true, tampl, tcent, twid, centerr, ww, arcnorm, nsig = arc.detect_lines(spec, sigdetect=sigdetec,
                                                                                  fwhm=fwhm, cont_samp=cont_samp,
                                                                                  debug=debug)

    peaks_good = tcent[ww]
    ampl_good = tampl[ww]
    # convert from pixel location to wavelength
    pixvec = np.arange(spec.size)
    wave_peak = scipy.interpolate.interp1d(pixvec, wavelength, bounds_error=False, fill_value='extrapolate')(peaks_good)
    # Vacuum?
    if vacuum:
        msgs.info("Writing wavelengths in vacuum")
        wave_peak = airtovac(wave_peak * units.AA).value

    npeak = len(wave_peak)
    ion = npeak*[str(line_name)]
    NIST = npeak*[1]
    Instr = npeak*[32]
    Source = npeak*['wavemodel.py']

    if iraf_frmt:
        msgs.info("Printing file in IRAF format: {}".format(file_root_name+'_iraf_lines.dat'))
        ion = np.array(ion)
        id_lines_iraf = np.vstack( (np.round(wave_peak,5), ion, np.round(ampl_good,5)) ).T
        np.savetxt(file_root_name+'_iraf_lines.dat', id_lines_iraf, fmt="%15s %6s %15s", delimiter="  ")
    else:
        msgs.info("Printing file: {}".format(file_root_name+'_lines.dat'))
        dat = Table([wave_peak, ion, NIST, Instr, ampl_good, Source], names=('wave', 'ion','NIST','Instr','amplitude','Source'))
        dat.write(file_root_name+'_lines.dat',format='ascii.fixed_width')


def create_OHlinelist(resolution, waveminmax=(0.8,2.6), dlam=40.0, flgd=True, nirsky_outfile=None,
                      fwhm=None, sigdetec=3., line_name='OH', file_root_name=None, iraf_frmt=False, 
                      debug=False):
    """Create a synthetic sky spectrum at a given resolution, extract significant lines, and
    store them in a PypeIt compatibile file. The skymodel is built from nearIR_modelsky and
    includes black body at 250K, OH lines, and H2O lines (but only at lambda>2.3microns).

    Parameters
    ----------
    resolution : np.float
        resolution of the spectrograph
    waveminmax : tuple
        wavelength range in microns to be covered by the model.
        Default is: (0.8, 2.6)
    dlam : 
        bin to be used to create the wavelength grid of the model.
        If flgd='True' it is a bin in velocity (km/s). If flgd='False'
        it is a bin in linear space (microns).
        Default is: 40.0 (with flgd='True')
    flgd : boolean
        if flgd='True' (default) wavelengths are created with 
        equal steps in log space. If 'False', wavelengths will be
        created wit equal steps in linear space.
    nirsky_outfile : str
        name of the fits file where the model sky spectrum will be stored.
        default is 'None' (i.e., no file will be written).
    fwhm : float
        fwhm in pixels used for filtering out arc lines that are too
        wide and not considered in fits. Parameter of arc.detect_lines().
        If set to 'None' the fwhm will be derived from the resolution as:
        2. * central_wavelength / resolution
    sigdetec : float
        sigma threshold above fluctuations for line detection. Parameter
        of arc.detect_lines(). Default = 2.
    line_name : str
        name of the lines to listed in the file. Default is 'OH'.
    file_root_name : str
        name of the file where the identified lines will be stored.
        The code automatically add '_lines.dat' at the end of the
        root name.
    iraf_frmt : bool
        if True, the file is written in the IRAF format (i.e. wavelength,
        ion name, amplitude).
    debug : boolean
        If True will show debug plots
    """

    wavelength, spec = nearIR_modelsky(resolution, waveminmax=waveminmax, dlam=dlam,
                                       flgd=flgd, nirsky_outfile=nirsky_outfile, debug=debug)

    if fwhm is None:
        msgs.warn("No min FWHM for the line detection set. Derived from the resolution at the center of the spectrum")
        wl_cent = np.average(wavelength)
        wl_fwhm = wl_cent / resolution
        wl_bin = np.abs((wavelength-np.roll(wavelength,1))[np.where(np.abs(wavelength-wl_cent)==np.min(np.abs(wavelength-wl_cent)))])
        # In order not to exclude all the lines, fwhm is set to 5 times
        # the minimum fwhm of the spectrum
        fwhm = 1.1 * wl_fwhm / wl_bin[0]
        if fwhm < 1.:
             msgs.warn("Lines are unresolved. Setting FWHM=2.pixels")
             fwhm = 2.

    if line_name is None:
        msgs.warn("No line_name as been set. The file will contain XXX as ion")
        line_name = 'XXX'

    if file_root_name is None:
        msgs.warn("No file_root_name as been set. The file will called OH_SKY_lines.dat")
        file_root_name = 'OH_SKY'

    create_linelist(wavelength, spec, fwhm=fwhm, sigdetec=sigdetec, line_name=line_name,
                    file_root_name=file_root_name, iraf_frmt=iraf_frmt, debug=debug)


def create_ThArlinelist(resolution, waveminmax=(3000.,10500.), dlam=40.0, flgd=True, thar_outfile=None,
                        fwhm=None, sigdetec=3., line_name='ThAr', file_root_name=None, iraf_frmt=False,
                        debug=False, vacuum=True):
    """Create a syntetic ThAr spectrum at a given resolution, extract significant lines, and
    store them in a PypeIt compatibile file. This is based on the Murphy et al. ThAr spectrum.
    Detailed information are here: http://astronomy.swin.edu.au/~mmurphy/thar/index.html

    Parameters
    ----------
    resolution : np.float
        resolution of the spectrograph
    waveminmax : tuple
        wavelength range in ang. to be covered by the model.
        Default is: (3000., 10500.)
    dlam : 
        bin to be used to create the wavelength grid of the model.
        If flgd='True' it is a bin in velocity (km/s). If flgd='False'
        it is a bin in linear space (angstrom).
        Default is: 40.0 (with flgd='True')
    flgd : boolean
        if flgd='True' (default) wavelengths are created with 
        equal steps in log space. If 'False', wavelengths will be
        created wit equal steps in linear space.
    thar_outfile : str
        name of the fits file where the model sky spectrum will be stored.
        default is 'None' (i.e., no file will be written).
    fwhm : float
        fwhm in pixels used for filtering out arc lines that are too
        wide and not considered in fits. Parameter of arc.detect_lines().
        If set to 'None' the fwhm will be derived from the resolution as:
        2. * central_wavelength / resolution
    sigdetec : float
        sigma threshold above fluctuations for line detection. Parameter
        of arc.detect_lines(). Default = 2.
    line_name : str
        name of the lines to listed in the file.
    file_root_name : str
        name of the file where the identified lines will be stored.
        The code automatically add '_lines.dat' at the end of the
        root name.
    iraf_frmt : bool
        if True, the file is written in the IRAF format (i.e. wavelength,
        ion name, amplitude).
    debug : boolean
        If True will show debug plots
    vacuum (bool):
        If True, write the wavelengths in vacuum
    """

    wavelength, spec = optical_modelThAr(resolution, waveminmax=waveminmax, dlam=dlam,
                                         flgd=flgd, thar_outfile=thar_outfile, debug=debug)
    if fwhm is None:
        msgs.warn("No min FWHM for the line detection set. Derived from the resolution at the center of the spectrum")
        wl_cent = np.average(wavelength)
        wl_fwhm = wl_cent / resolution
        wl_bin = np.abs((wavelength-np.roll(wavelength,1))[np.where(np.abs(wavelength-wl_cent)==np.min(np.abs(wavelength-wl_cent)))])
        # In order not to exclude all the lines, fwhm is set to 5 times
        # the minimum fwhm of the spectrum
        fwhm = 1.1 * wl_fwhm / wl_bin[0]
        if fwhm < 1.:
             msgs.warn("Lines are unresolved. Setting FWHM=2.*pixels")
             fwhm = 2.

    if line_name is None:
        msgs.warn("No line_name as been set. The file will contain XXX as ion")
        line_name = 'XXX'

    if file_root_name is None:
        msgs.warn("No file_root_name as been set. The file will called ThAr_lines.dat")
        file_root_name = 'ThAr'

    create_linelist(wavelength, spec, fwhm=fwhm, sigdetec=sigdetec, line_name=line_name,
                    file_root_name=file_root_name, iraf_frmt=iraf_frmt, debug=debug, vacuum=vacuum)





"""
Coadding module.

.. include common links, assuming primary doc root is up one directory
.. include:: ../links.rst
"""

import os

from IPython import embed

import numpy as np
from astropy.io import fits
from pypeit.spectrographs.util import load_spectrograph
from pypeit import utils
from pypeit import specobjs
from pypeit import msgs
from pypeit.core import coadd


class CoAdd1D(object):

    @classmethod
    def get_instance(cls, spec1dfiles, objids, par=None, sensfile=None, debug=False, show=False):
        """
        Superclass factory method which generates the subclass instance. See __init__ docs for arguments.
        """
        pypeline = fits.getheader(spec1dfiles[0])['PYPELINE'] + 'CoAdd1D'
        return next(c for c in cls.__subclasses__() if c.__name__ == pypeline)(
            spec1dfiles, objids, par=par, sensfile=sensfile, debug=debug, show=show)

    def __init__(self, spec1dfiles, objids, par=None, sensfile=None, debug=False, show=False):
        """

        Args:
            spec1dfiles (list):
               List of strings which are the spec1dfiles
            objids (list):
               List of strings which are the objids for the object in each spec1d file that you want to coadd
            par (parset):
               Pypeit parameter set object
            sensfile (str): optional
               File holding the sensitivity function. This is required for echelle coadds only.
            debug (bool): optional
               Debug. Default = False
            show (bool):
               Debug. Default = True
        """
        # Instantiate attributes
        self.spec1dfiles = spec1dfiles
        self.objids = objids
        self.sensfile = sensfile
        # Load the spectrograph only if par was not passed in to get default parset
        if par is None:
            header = fits.getheader(spec1dfiles[0])
            spectrograph = load_spectrograph(header['PYP_SPEC'])
            self.par = spectrograph.default_pypeit_par()['coadd1d']
        else:
            self.par = par
        self.debug = debug
        self.show = show
        self.nexp = len(self.spec1dfiles) # Number of exposures
        self.coaddfile = None

    def run(self):
        """
        Runs the coadding
        """

        # Load the data
        self.waves, self.fluxes, self.ivars, self.masks, self.header = self.load()
        # Coadd the data
        self.wave_coadd, self.flux_coadd, self.ivar_coadd, self.mask_coadd = self.coadd()

    def load(self):
        """
        Load the arrays we need for performing coadds.

        Returns:
            tuple:
               - waves, fluxes, ivars, masks, header
        """

        for iexp in range(self.nexp):
            sobjs = specobjs.SpecObjs.from_fitsfile(self.spec1dfiles[iexp])
            indx = sobjs.name_indices(self.objids[iexp])
            if not np.any(indx):
                msgs.error("No matching objects for {:s}.  Odds are you input the wrong OBJID".format(self.objids[iexp]))
            wave_iexp, flux_iexp, ivar_iexp, mask_iexp, meta_spec, header = \
                sobjs[indx].unpack_object(ret_flam=self.par['flux_value'])
            # Allocate arrays on first iteration
            if iexp == 0:
                waves = np.zeros(wave_iexp.shape + (self.nexp,))
                fluxes = np.zeros_like(waves)
                ivars = np.zeros_like(waves)
                masks = np.zeros_like(waves, dtype=bool)
                header_out = header

            waves[...,iexp], fluxes[...,iexp], ivars[..., iexp], masks[...,iexp] = wave_iexp, flux_iexp, ivar_iexp, mask_iexp

        return waves, fluxes, ivars, masks, header_out

    def save(self, coaddfile, telluric=None, obj_model=None, overwrite=True):
        """
        Routine to save 1d coadds to a fits file. This replaces save.save_coadd1d_to_fits

        Args:
            coaddfile (str):
               File to outuput coadded spectrum to.
            telluric (str):
               This is vestigial and should probably be removed.
            obj_model (str):
               This is vestigial and should probably be removed
            overwrite (bool):
               Overwrite existing file?

        """

        self.coaddfile = coaddfile
        ex_value = self.par['ex_value']
        # Estimate sigma from ivar
        sig = np.sqrt(utils.inverse(self.ivar_coadd))
        if (os.path.exists(self.coaddfile)) and (np.invert(overwrite)):
            hdulist = fits.open(self.coaddfile)
            msgs.info("Reading primary HDU from existing file: {:s}".format(self.coaddfile))
        else:
            msgs.info("Creating a new primary HDU.")
            prihdu = fits.PrimaryHDU()
            if self.header is None:
                msgs.warn('The primary header is none')
            else:
                prihdu.header = self.header
            hdulist = fits.HDUList([prihdu])

        wave_mask = self.wave_coadd > 1.0
        # Add Spectrum Table
        cols = []
        cols += [fits.Column(array=self.wave_coadd[wave_mask], name='{:}_WAVE'.format(ex_value), format='D')]
        cols += [fits.Column(array=self.flux_coadd[wave_mask], name='{:}_FLAM'.format(ex_value), format='D')]
        cols += [fits.Column(array=self.ivar_coadd[wave_mask], name='{:}_FLAM_IVAR'.format(ex_value), format='D')]
        cols += [fits.Column(array=sig[wave_mask], name='{:}_FLAM_SIG'.format(ex_value), format='D')]
        cols += [fits.Column(array=self.mask_coadd[wave_mask].astype(float), name='{:}_MASK'.format(ex_value), format='D')]
        if telluric is not None:
            cols += [fits.Column(array=telluric[wave_mask], name='TELLURIC', format='D')]
        if obj_model is not None:
            cols += [fits.Column(array=obj_model[wave_mask], name='OBJ_MODEL', format='D')]

        coldefs = fits.ColDefs(cols)
        tbhdu = fits.BinTableHDU.from_columns(coldefs)
        tbhdu.name = 'OBJ0001-SPEC0001-{:}'.format(ex_value.capitalize())
        hdulist.append(tbhdu)

        if (os.path.exists(self.coaddfile)) and (np.invert(overwrite)):
            hdulist.writeto(self.coaddfile, overwrite=True)
            msgs.info("Appending 1D spectra to existing file {:s}".format(self.coaddfile))
        else:
            hdulist.writeto(self.coaddfile, overwrite=overwrite)
            msgs.info("Wrote 1D spectra to {:s}".format(self.coaddfile))



    def coadd(self):
        """
        Dummy method overloaded by sub-classes

        Returns:

        """
        return (None,)*4


class MultiSlitCoAdd1D(CoAdd1D):
    """
    Child of CoAdd1d for Multislit and Longslit reductions
    """

    def __init__(self, spec1dfiles, objids, par=None, sensfile=None, debug=False, show=False):
        """

            Args:
                spec1dfiles (list):
                   List of strings which are the spec1dfiles
                objids (list):
                   List of strings which are the objids for the object in each spec1d file that you want to coadd
                par (parset):
                   Pypeit parameter set object
                sensfile (str): optional
                   File holding the sensitivity function. This is required for echelle coadds only.
                debug (bool): optional
                   Debug. Default = False
                show (bool):
                   Debug. Default = True
        """

        super().__init__(spec1dfiles, objids, par=par, sensfile=sensfile, debug=debug, show=show)


    def coadd(self):
        """
        Perform coadd for for Multi/Longslit data using multi_combspec

        Returns:
            tuple
              - wave, flux, ivar, mask

        """
        return coadd.multi_combspec(
            self.waves, self.fluxes, self.ivars, self.masks,
            sn_smooth_npix=self.par['sn_smooth_npix'], wave_method=self.par['wave_method'],
            samp_fact=self.par['samp_fact'], ref_percentile=self.par['ref_percentile'],
            maxiter_scale=self.par['maxiter_scale'], sigrej_scale=self.par['sigrej_scale'],
            scale_method=self.par['scale_method'], sn_min_medscale=self.par['sn_min_medscale'],
            sn_min_polyscale=self.par['sn_min_polyscale'], maxiter_reject=self.par['maxiter_reject'],
            lower=self.par['lower'], upper=self.par['upper'], maxrej=self.par['maxrej'], sn_clip=self.par['sn_clip'],
            debug=self.debug, show=self.show)





class EchelleCoAdd1D(CoAdd1D):
    """
    Child of CoAdd1d for Echelle reductions
    """

    def __init__(self, spec1dfiles, objids, par=None, sensfile=None, debug=False, show=False):
        """

            Args:
                spec1dfiles (list):
                   List of strings which are the spec1dfiles
                objids (list):
                   List of strings which are the objids for the object in each spec1d file that you want to coadd
                par (parset):
                   Pypeit parameter set object
                sensfile (str): optional
                   File holding the sensitivity function. This is required for echelle coadds only.
                debug (bool): optional
                   Debug. Default = False
                show (bool):
                   Debug. Default = True
        """

        super().__init__(spec1dfiles, objids, par=par, sensfile=sensfile, debug=debug, show=show)

    def coadd(self):
        """
        Perform coadd for for echelle data using ech_combspec

        Returns:
            tuple
              - wave, flux, ivar, mask

        """
        (wave_coadd, flux_coadd, ivar_coadd, mask_coadd), order_stacks = coadd.ech_combspec(
            self.waves, self.fluxes, self.ivars, self.masks, self.sensfile,
            nbest=self.par['nbest'], sn_smooth_npix=self.par['sn_smooth_npix'], wave_method=self.par['wave_method'],
            samp_fact=self.par['samp_fact'], ref_percentile=self.par['ref_percentile'],
            maxiter_scale=self.par['maxiter_scale'], sigrej_scale=self.par['sigrej_scale'],
            scale_method=self.par['scale_method'], sn_min_medscale=self.par['sn_min_medscale'],
            sn_min_polyscale=self.par['sn_min_polyscale'], maxiter_reject=self.par['maxiter_reject'],
            lower=self.par['lower'], upper=self.par['upper'], maxrej=self.par['maxrej'], sn_clip=self.par['sn_clip'],
            debug = self.debug, show = self.show)

        return wave_coadd, flux_coadd, ivar_coadd, mask_coadd
